"""Unit tests related to duplex.image module."""
import tempfile
from os.path import abspath
from unittest import TestCase
from unittest.mock import patch
from base64 import b64encode

import pytest
import numpy
import tensorflow
from tensorflow import io as io_ops

from pupyl.duplex.image import ImageIO
from pupyl.duplex.exceptions import FileIsNotImage


TEST_LOCAL = abspath('tests/test_image.jpg')
TEST_LOCAL_RBGA = abspath('tests/test_png.png')
TEST_LOCAL_GRAY = abspath('tests/test_grayscale.jpg')
TEST_LOCAL_GIF = abspath('tests/test_gif.gif')
TEST_NOT_IMAGE = abspath('tests/not_image.jpg')
TEST_SAVED_TENSOR = abspath('tests/compressed_image.npy')
TEST_SIZE = (280, 260)
TEST_RESIZE = (100, 100)


class TestCases(TestCase):
    """Unit tests over special cases."""

    def test_get_image_no_image(self):
        """Unit test for get_image method, not image case."""
        with self.assertRaises(FileIsNotImage):
            ImageIO.get_image(TEST_NOT_IMAGE)

    def test_get_image_bytes_to_base64_no_image(self):
        """Unit test to method get_image_bytes_to_base64."""
        with self.assertRaises(FileIsNotImage):
            with open(TEST_NOT_IMAGE, 'rb') as image_bytes:
                ImageIO.get_image_bytes_to_base64(image_bytes.read())


def test_get_image_as_tensor():
    """Unit test for get_image method, as_tensor optional parameter set."""
    assert isinstance(
        ImageIO.get_image(TEST_LOCAL, as_tensor=True),
        tensorflow.Tensor
    )

    assert isinstance(
        ImageIO.get_image(TEST_LOCAL_GRAY, as_tensor=True),
        tensorflow.Tensor
    )

    assert isinstance(
        ImageIO.get_image(TEST_LOCAL_RBGA, as_tensor=True),
        tensorflow.Tensor
    )

    assert isinstance(
        ImageIO.get_image(TEST_LOCAL_GIF, as_tensor=True),
        tensorflow.Tensor
    )


def test_mean_gif():
    """Unit test for mean_gif method."""
    assert ImageIO.mean_gif(TEST_LOCAL_GIF).ndim == 3


def test_mean_gif_not_gif():
    """Unit test for mean_gif method, not a gif file."""
    assert ImageIO.mean_gif(TEST_LOCAL).ndim == 3


def test_size_size():
    """Unit test for size method, return size case."""
    assert ImageIO.size(TEST_LOCAL) == TEST_SIZE


def test_size_resize():
    """
    Unit test for size method, return size case
    """
    assert ImageIO.size(TEST_LOCAL, TEST_RESIZE).shape[:2] == TEST_RESIZE


def test_compress():
    """
    Unit test for compress method
    """
    def behaviour_compress(tensor):
        """
        Closure with expected behaviour for compress
        method
        """
        return io_ops.encode_jpeg(
            tensor,
            quality=80,
            progressive=True,
            optimize_size=True,
            chroma_downsampling=True
        )

    test_image_bytes = ImageIO.get_image(TEST_LOCAL)
    test_tensor = ImageIO.encoded_to_tensor(test_image_bytes)

    numpy.testing.assert_array_equal(
        ImageIO.compress(test_tensor),
        behaviour_compress(test_tensor)
    )

    numpy.testing.assert_array_equal(
        ImageIO.compress(test_tensor, as_tensor=True),
        ImageIO.encoded_to_tensor(behaviour_compress(test_tensor))
    )


def test_encoded_to_compressed_tensor():
    """Unit test for encoded_to_compressed_tensor method."""
    saved_tensor = numpy.load(TEST_SAVED_TENSOR)
    test_tensor = ImageIO.encoded_to_compressed_tensor(
        ImageIO.get_image(TEST_LOCAL)
    )

    assert test_tensor.ndim == 1 and test_tensor.shape == 6104
    numpy.testing.assert_array_equal(saved_tensor, test_tensor)


def test_save_image():
    """Unit test for save_image method."""
    test_tensor_bytes = ImageIO.get_image(TEST_LOCAL)

    with tempfile.NamedTemporaryFile() as temp_file:
        ImageIO.save_image(temp_file.name, test_tensor_bytes)

        assert test_tensor_bytes == ImageIO.get_image(temp_file.name)


def test_get_image_base64():
    """Unit test for method get_image_base64."""
    with open(TEST_LOCAL, 'rb') as image_bytes:
        image_bytes = image_bytes.read()
        test_image_b64 = b64encode(image_bytes)

    assert test_image_b64 == ImageIO.get_image_base64(TEST_LOCAL)


def test_get_image_bytes_to_base64():
    """Unit test for method get_image_bytes_to_base64."""
    with open(TEST_LOCAL, 'rb') as image_bytes:
        image_bytes = image_bytes.read()
        test_image_b64 = b64encode(image_bytes)

    assert test_image_b64 == ImageIO.get_image_bytes_to_base64(image_bytes)


@pytest.mark.parametrize(
    'os, expected_support',
    [
        ('Windows', False),
        ('Linux', True),
        ('Darwin', True)
    ]
)
@patch('pupyl.duplex.image.system')
def test_os_supports_compression(system_mock, os, expected_support):
    """Unit test for method os_supports_compression."""
    system_mock.return_value = os

    real_support = ImageIO.os_supports_compression()

    assert real_support == expected_support



"""
🧿 pupyl

Pupyl is a really fast image search library which you
can index your own (millions of) images and find similar
images in milliseconds.
"""
__version__ = 'v0.14.5'


import os
import json
from enum import Enum
import concurrent.futures

from pupyl.indexer.facets import Index
from pupyl.duplex.file_io import FileIO
from pupyl.storage.database import ImageDatabase
from pupyl.duplex.exceptions import FileIsNotImage
from pupyl.embeddings.features import Extractors, Characteristics


class PupylImageSearch:
    """Encapsulates every aspects of ``pupyl``, from feature extraction
    to indexing and image storaging.
    """

    def __init__(
            self,
            data_dir=None,
            extreme_mode=True,
            **kwargs
    ):
        """Pupyl image search factory.

        Parameters
        ----------
        data_dir: str
            The directory where all assets are stored.

        extreme_mode: bool
            If should the extreme mode (faster execution but more memory
            consumption) be enabled or not.

        import_images: bool
            If images should (or was) imported into the database.

        characteristic: Enum or int or str
            The characteristic for feature extraction that must be used. If
            reading from an already created database, retrieves it from the
            (internal) configuration files. It supports retrieval by the
            ``Characteristics`` (``enum``), by its ``name`` (``str``) or
            by its ``value`` (``int``). For more information, see
            :ref:`features:📊 features module`.

        Notes
        -----
        A ``characteristic`` defines a feature extractor, with its own
        complexity and balance between search precision and indexing speed.
        Below are the description of every possible
        ``characteristic``,
        indexing the dataset
        `<https://github.com/policratus/pupyl/raw/main/samples/images.tar.xz>`_,
        containing 594 images.

        .. list-table::
           :header-rows: 1

           * - Name
             - Value
             - Network
             - Speed (GPU)
           * - MINIMUMWEIGHT_FAST_SMALL_PRECISION
             - 1
             - MobileNet
             - 19 s
           * - LIGHTWEIGHT_FAST_SMALL_PRECISION
             - 2
             - ResNet50V2
             - 21.2 s
           * - LIGHTWEIGHT_FAST_SHORT_PRECISION
             - 3
             - ResNet101V2
             - 20.3 s
           * - LIGHTWEIGHT_QUICK_SHORT_PRECISION
             - 4
             - DenseNet169
             - 32.7 s
           * - MEDIUMWEIGHT_QUICK_GOOD_PRECISION
             - 5
             - DenseNet201
             - 31.2 s
           * - MIDDLEWEIGHT_QUICK_GOOD_PRECISION
             - 6
             - InceptionV3
             - 27 s
           * - MIDDLEWEIGHT_SLOW_GOOD_PRECISION
             - 7
             - Xception
             - 20.2 s
           * - HEAVYWEIGHT_SLOW_GOOD_PRECISION
             - 8
             - EfficientNetV2M
             - 39.3 s
           * - HEAVYWEIGHT_SLOW_HUGE_PRECISION
             - 9
             - EfficientNetV2L
             - 1min 1s

        *All the tests above were under the same circumstances and resources.*

        By default, ``pupyl`` chooses ``MINIMUMWEIGHT_FAST_SMALL_PRECISION``,
        the fastest but with not so much accuracy than
        'HEAVYWEIGHT_SLOW_HUGE_PRECISION', for instance.

        Examples
        --------
        from pupyl.search import PupylImageSearch

        # Creating a database using the extractor number 3,

        # LIGHTWEIGHT_FAST_SHORT_PRECISION, network ResNet101V2

        pupyl = PupylImageSearch(characteristic=3)

        # Creating a database with extractor 'HEAVYWEIGHT_SLOW_GOOD_PRECISION',

        # based on EfficientNetV2M

        characteristic = 'HEAVYWEIGHT_SLOW_GOOD_PRECISION'

        pupyl = PupylImageSearch(characteristic=characteristic)
        """
        self._extreme_mode = extreme_mode

        if data_dir:
            self._data_dir = data_dir
        else:
            self._data_dir = FileIO.pupyl_temp_data_dir()

        self._index_config_path = os.path.join(self._data_dir, 'index.json')

        configurations = self._index_configuration('r')

        if configurations:
            self._import_images = configurations['import_images']
            self._characteristic = Characteristics.by_name(
                configurations['characteristic']
            )

            if configurations.get('feature_size'):
                self._feature_size = configurations['feature_size']
        else:
            import_images = kwargs.get('import_images')
            characteristic = kwargs.get('characteristic')

            if import_images:
                self._import_images = import_images
            else:
                self._import_images = True

            if isinstance(characteristic, Enum):
                self._characteristic = characteristic
            elif isinstance(characteristic, int):
                self._characteristic = Characteristics.by_value(
                    characteristic
                )
            elif isinstance(characteristic, str):
                self._characteristic = Characteristics.by_name(
                    characteristic
                )
            else:
                self._characteristic = Characteristics.\
                    MINIMUMWEIGHT_FAST_SMALL_PRECISION

        self.image_database = ImageDatabase(
            import_images=self._import_images,
            data_dir=self._data_dir
        )

        if extreme_mode:
            self.extractor = Extractors(self._characteristic)

            self.indexer = Index(
                self.extractor.output_shape,
                data_dir=self._data_dir
            )

    def _index_configuration(self, mode, **kwargs):
        """Loads or saves an index configuration file, if exists.

        Parameters
        ----------
        mode: str
            Defines which mode should be used over the configuration
            file. 'r' is for file reading, 'w' for writing.

        feature_size: int
            The size of the current feature extraction method.

        Returns
        -------
        dict or bool:
            Returns a ``dict`` if an already saved database are found,
            containing several database configurations or ``bool`` ``True``
            if a new configuration file was created, or ``bool`` ``False`` if
            either a configuration couldn't be created or loaded.
        """
        try:
            with open(
                self._index_config_path, mode, encoding='utf-8'
            ) as config_file:
                if mode == 'r':

                    return json.load(config_file)

                if mode == 'w':
                    feature_size = kwargs.get('feature_size')

                    configurations = {
                        'import_images': self._import_images,
                        'characteristic': self._characteristic.name,
                    }

                    if feature_size:
                        configurations['feature_size'] = feature_size

                    json.dump(configurations, config_file)

                return True
        except FileNotFoundError:
            return False

    def index(self, uri, **kwargs):
        """Performs parallel image indexing.

        Parameters
        ----------
        uri: str
            Directory or file, or http(s) location.

        check_unique: bool
            If, during the index process, imported images
            should have their unicity verified (to avoid duplicates).

        Attention
        ---------
        If ``check_unique`` is ``True``, consequentely unicity checks will be
        performed, which creates some overheads on the index process, making
        it slower.
        """
        check_unique = kwargs.get('check_unique')

        if check_unique is None:
            check_unique = False

        if self._extreme_mode:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = {
                    executor.submit(
                        self.image_database.insert,
                        rank,
                        uri_from_file
                    ): rank
                    for rank, uri_from_file in self.extractor.progress(
                        enumerate(
                            self.extractor.scan_images(uri)
                        ),
                        precise=False,
                        message='Importing images:'
                    )
                }

                ranks = []

                for future in self.extractor.progress(
                    concurrent.futures.as_completed(futures),
                    precise=False,
                    message='Scanning images:'
                ):
                    ranks.append(futures[future])

            with concurrent.futures.ThreadPoolExecutor() as executor:
                try:
                    futures = {
                        executor.submit(
                            self.extractor.extract_save,
                            self.image_database.mount_file_name(
                                rank, extension='npy'
                            ),
                            self.image_database.load_image_metadata(
                                rank, filtered=['internal_path']
                            )['internal_path']
                        ): rank
                        for rank in ranks
                    }

                    _ = [
                        *self.extractor.progress(
                            concurrent.futures.as_completed(futures),
                            precise=False,
                            message='Extracting features:'
                        )
                    ]

                except IndexError as index_error:
                    raise FileIsNotImage('Please, check your input images.') \
                        from index_error

            for rank in self.extractor.progress(
                sorted(ranks),
                precise=True,
                message='Indexing features:'
            ):
                self.indexer.append(
                    self.extractor.load(
                        self.image_database.mount_file_name(
                            rank, extension='npy'
                        )
                    ),
                    check_unique=check_unique
                )

            self.indexer.flush()

            with concurrent.futures.ThreadPoolExecutor() as executor:
                for rank in ranks:
                    executor.submit(
                        self.indexer.remove_feature_cache, rank
                    )

            self._index_configuration(
                'w', feature_size=self.extractor.output_shape
            )

        else:
            with Extractors(
                characteristics=self._characteristic,
                extreme_mode=self._extreme_mode
            ) as extractor:
                with Index(
                    extractor.output_shape, data_dir=self._data_dir
                ) as indexer:
                    for rank, uri_from_file in extractor.progress(
                        enumerate(
                            extractor.scan_images(uri)
                        ),
                        precise=False,
                        message='Indexing images:'
                    ):
                        self.image_database.insert(rank, uri_from_file)

                        extractor.extract_save(
                            self.image_database.mount_file_name(
                                rank, extension='npy'
                            ),
                            self.image_database.load_image_metadata(
                                rank, filterd=['internal_path']
                            )['internal_path']
                        )

                        indexer.append(
                            extractor.load(
                                self.image_database.mount_file_name(
                                    rank, extension='npy'
                                )
                            ),
                            check_unique=check_unique)

                    self._index_configuration(
                        'w', feature_size=extractor.output_shape
                    )

    def search(
        self,
        query,
        top=4,
        return_metadata=False,
        return_distances=False
    ):
        """Executes the search for a similar image throughout the database
        based on the ``query`` image.

        Parameters
        ----------
        query: str
            URI of a image to be used as query.

        top: int (optional)(default: 4)
            How many results should be returned from the search process.

        return_metadata: bool (optional)(default: False)
            If the image results metadata should also be returned.

        return_distances: bool (optional)(default: False)
            If the method should return the distances between the ``query``
            image and other images present in the database.

        Yields
        ------
        int, dict or tuple:
            Respectively describing the image index identification that is
            decresingly (ordered) similar from the query image, a ``dict``
            with metadata information about this images (case when
            ``return_metadata=True``) or a ``tuple`` when the method
            was asked to return the distances but not the image metadata
            (case when ``return_metadata=False`` and ``return_distances=True``.
        """
        if self._extreme_mode:
            for result in self.indexer.search(
                self.extractor.extract(query),
                results=top,
                return_distances=return_distances
            ):
                if return_distances:
                    yield self.image_database.load_image_metadata(
                        result[0], distance=result[1]
                    ) if return_metadata else result
                else:
                    yield self.image_database.load_image_metadata(result) \
                        if return_metadata else result
        else:
            with Extractors(
                characteristics=self._characteristic,
                extreme_mode=self._extreme_mode
            ) as extractor:
                with Index(
                    extractor.output_shape, data_dir=self._data_dir
                ) as indexer:
                    for result in indexer.search(
                        extractor.extract(query),
                        results=top,
                        return_distances=return_distances
                    ):
                        if return_distances:
                            yield self.image_database.load_image_metadata(
                                result[0], distance=result[1]
                            ) if return_metadata else result
                        else:
                            yield self.image_database.load_image_metadata(
                                result
                            ) if return_metadata else result

    def remove(self, index):
        """Removes an indexed image from the storage.

        Parameters
        ----------
        index: int
            The image to be deleted, based on ``index``
            (internal image identification).

        Example
        -------
        search.remove(12) # Will remove image with ``index`` 12 from the
        storage.
        """
        if self._extreme_mode:
            self.indexer.remove(index)
        else:
            with Extractors(
                characteristics=self._characteristic,
                extreme_mode=self._extreme_mode
            ) as extractor:
                with Index(
                    extractor.output_shape, data_dir=self._data_dir
                ) as indexer:
                    indexer.remove(index)

        self.image_database.remove(index)



"""Operations of file and mime type discovery, based on file content."""

from enum import Enum

from pupyl.duplex import exceptions


class FileHeaderHex(Enum):
    """Describes header hexadecimals for every supported file types.

    The current ones supported are:

    ``JPG, GIF, PNG, TIF, GZP, BZ2, ZIP, LXZ``"""

    JPG = 'ffd8ff'
    GIF = '474946'
    PNG = '89504e'
    TIF = '49492a'
    GZP = '1f8b08'
    BZ2 = '425a68'
    ZIP = '504b03'
    LXZ = 'fd377a'


class FileMimeTypes(Enum):
    """Describes the mime types associated with underlying files.

    The current ones supported are:

    ``JPG, GIF, PNG, TIF, GZP, BZ2, ZIP, LXZ``"""

    JPG = 'image/jpeg'
    GIF = 'image/gif'
    PNG = 'image/png'
    TIF = 'image/tiff'
    GZP = 'application/gzip'
    BZ2 = 'application/x-bzip2'
    ZIP = 'application/zip'
    LXZ = 'application/x-xz'
    TXZ = 'application/tar-xz'
    TGZ = 'application/tar-gz'
    TZ2 = 'application/tar-bz2'


class TarCompressedTypes:
    """Describes all supported tar compressed types.

    The current ones supported are:

    ``GZIP, LZMA, BZIP2``"""
    _types = {
        'gzip': {
            'mimetype': FileMimeTypes.TGZ.value,
            'name': FileMimeTypes.TGZ.name
        },
        'xz': {
            'mimetype': FileMimeTypes.TXZ.value,
            'name': FileMimeTypes.TXZ.name
        },
        'bzip2': {
            'mimetype': FileMimeTypes.TZ2.value,
            'name': FileMimeTypes.TZ2.name
        }
    }

    @classmethod
    def name(cls, sub_type):
        """Returns the name of some sub (file) type.

        Parameters
        ----------
        sub_type: str
            The name of one subtype.

        Returns
        -------
        str:
            With the found subtype.

        Example
        -------
        ``TarCompressedTypes.name('xz') # Returns 'TXZ'``
        """
        return cls._types[sub_type]['name']

    @classmethod
    def mime(cls, sub_type):
        """Return the mime type of some sub type.

        Parameters
        ----------
        sub_type: str
            The name of one subtype.

        Returns
        -------
        str:
            With the found subtype mimetype.

        Example
        -------
        ``TarCompressedTypes.name('xz') # Returns 'application/tar-xz'``
        """
        return cls._types[sub_type]['mimetype']


class FileType:
    """Operations of file type discovery over file content."""

    @staticmethod
    def to_hex(bytess):
        """
        Converts a byte string to its hexadecimal representation.

        Parameters
        ----------
        bytess: bytes
            Bytes of some file

        Returns
        -------
        str:
            With the hexadecimal representation
        """
        return bytess.hex()

    @classmethod
    def header(cls, bytess):
        """
        Return only the file header or the first three bytes.

        Parameters
        ----------
        bytess: bytes
            Bytes of some file.

        Returns
        -------
        str:
            With the header hexadecimal representation.
        """
        return cls.to_hex(bytess[:3])

    @classmethod
    def guess_file_type(cls, bytess, mimetype=False):
        """Heuristically discover the file type for the underlying path.

        Parameters
        ----------
        bytess: bytes
            Bytes of some file.

        mimetype: bool
            If the mimetype should be returned or not.

        Raises
        ------
        FileTypeNotSupportedYet:
            For an unsupported file.

        Returns
        -------
        str:
            With inferred type.
        """
        try:
            header = cls.header(bytess)
        except TypeError as type_error:
            raise exceptions.FileTypeNotSupportedYet from type_error

        for hexa in FileHeaderHex:
            if hexa.value == header:
                if mimetype:

                    return hexa, FileMimeTypes.__members__.get(hexa.name).value

                return hexa.name

        raise exceptions.FileTypeNotSupportedYet(
            'Input file type is not supported yet.'
        )

    @staticmethod
    def tar_compressed_types_resolve(sub_type, mimetype=False):
        """Resolves a tar file compression type (if any).

        Parameters
        ----------
        sub_type: str
            The sub type of the tar compressed type.

        mimetype: bool
            If should be returned the MIME type instead of internal
            file type name.

        Returns
        -------
        str:
            With inferred tar (compressed) type
        """
        return TarCompressedTypes.mime(sub_type) \
            if mimetype else TarCompressedTypes.name(sub_type)

    @classmethod
    def is_image(cls, bytess):
        """Return if the read file is an image or not.

        Parameters
        ----------
        bytess: bytes
            Bytes of some file

        Raises
        ------
        FileTypeNotSupportedYet:
            For an unsupported file.

        Returns
        -------
        bool:
            Describing if the inferred type is image
        """
        try:
            return cls.guess_file_type(
                bytess,
                mimetype=True
            )[1].split('/')[0] == 'image'
        except exceptions.FileTypeNotSupportedYet:
            return False



"""Operations over files, introspection and more."""

import os
import csv
import bz2
import lzma
import gzip
import uuid
import tarfile
import zipfile
import tempfile
import mimetypes
from io import BytesIO
from itertools import cycle
from enum import Enum, auto
from datetime import datetime
from urllib.parse import urlparse
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

import termcolor

from pupyl.duplex.file_types import FileType
from pupyl.duplex.exceptions import FileTypeNotSupportedYet, \
    FileScanNotPossible
from pupyl.addendum.operators import intmul


class Protocols(Enum):
    """Defines several possible protocol enumerators to be discovered.

    Notes
    -----
    The current supported protocols are:

    ``UNKNOWN: Unknown protocol``

    ``HTTP: Hypertext Transfer Protocol (also Secure supported)``

    ``FILE: Local storage file.``
    """

    UNKNOWN = auto()
    HTTP = auto()
    FILE = auto()


class FileIO(FileType):
    """Operations over files.

    Handling operations like temporary directories and files, retrieval of
    remote or local files, progress bars, file metadata, among others.
    """

    # Maximum file size supported, in GBs
    max_file_size = 8

    @staticmethod
    def pupyl_temp_data_dir():
        """Returns the path of a temporary directory to store pupyl assets.

        Returns
        -------
        str:
            A path containing the underlying temporary directory, found in the
            current operating system, added with a special directory for
            saving pupyl assets.
        """
        return os.path.join(
            tempfile.gettempdir(),
            'pupyl'
        )

    @classmethod
    def _get_url(cls, url, **kwargs):
        """Loads a file from a remote (http(s)) location.

        Parameters
        ----------
        url: str
            The URL where the image are stored.

        headers: dict
            A header to be passed through the HTTP request. Usually
            contains a header with an user-agent defined, like
            ``{'User-Agent': 'Mozilla/5.0'}``

        info: bool
            Defines if should be returned metadata information from
            the ``url``, instead of its ``bytes``.

        retry: int
            Counter for the number of retries already issued.

        Returns
        -------
        bytes or http.client.HTTPMessage:
            ``bytes`` with image binary information or
            ``http.client.HTTPMessage`` with file information (case
            ``info`` is ``True``).
        """
        max_retries = 3

        if kwargs.get('headers'):
            request = Request(url, headers=kwargs['headers'])
        else:
            request = url

        if kwargs.get('retry'):
            retry = kwargs['retry']
        else:
            retry = 1

        try:
            with urlopen(request, timeout=1) as ffile:
                if kwargs.get('info'):
                    return ffile.info()

                content_length = ffile.info().get_all('Content-Length')

                if content_length:
                    if int(
                        int(content_length[0]) / (2 ** 30)
                    ) >= cls.max_file_size:
                        print(
                            f'File {url} is bigger than {cls.max_file_size} '
                            ' GB. Cowardly refusing to read it.'
                        )

                        return b''

                return ffile.read()

        except HTTPError as http_error:
            print(
                f'URL {url} returned HTTP error {http_error.code}, '
                f'"{http_error.reason}". Retrying using other methods.'
            )

            if http_error.code == 403:
                return cls._get_url(
                    url, headers={'User-Agent': 'Mozilla/5.0'}
                )

            return b''

        except URLError as url_error:
            print(
                f'URL {url} request has thrown an error: {url_error.reason}. '
                f'Retrying: {retry} of {max_retries}.'
            )

            retry += 1

            if retry <= max_retries:
                return cls._get_url(url, retry=retry)

            raise URLError(f'URL {url} exhausted all retries. Giving up.') \
                from url_error

    @classmethod
    def _get_local(cls, path):
        """Loads a local file returning its bytes.

        Parameters
        ----------
        path: str
            Location which the file is saved.

        Returns
        -------
        bytes:
            With file binary information contained on the file.
        """
        if int(os.path.getsize(path) / (2 ** 30)) >= cls.max_file_size:
            print(
                f'File {path} is bigger than {cls.max_file_size} GB. '
                'Cowardly refusing to read it.'
            )
        else:
            with open(path, 'rb') as ffile:
                return ffile.read()

        return b''

    @classmethod
    def get(cls, uri):
        """Loads a file from specified location, remote or local.

        Parameters
        ----------
        uri: str
            Location where the file are stored.

        Returns
        -------
        bytes or Enum:
            If successful, returns the file bytes,
            or an Enum describing that the format wasn't recognized.
        """
        if cls._infer_protocol(uri) is Protocols.FILE:
            if urlparse(uri).scheme == 'file':
                uri = cls._file_scheme_to_path(uri)

            return cls._get_local(uri)

        if cls._infer_protocol(uri) is Protocols.HTTP:
            return cls._get_url(uri)

        return Protocols.UNKNOWN

    @staticmethod
    def _file_scheme_to_path(uri):
        """Converter from a file:// scheme to a path.

        Parameters
        ----------
        uri: str
            An URI to convert from `file://` scheme to a path

        Example
        -------
        ``FileIO._file_scheme_to_path(file:///home/policratus/1073140.jpg)``
        ``# Returns '/home/policratus/1073140.jpg'``
        """
        return uri[len('file://'):]

    @classmethod
    def get_metadata(cls, uri):
        """Returns underlying file metadata.

        Parameters
        ----------
        uri: str
            Location where the file are stored.

        Returns
        -------
        dict:
            Describing several file metadata
        """
        if cls._infer_protocol(uri) is Protocols.FILE:
            file_statistics = os.stat(uri)

            original_path, original_file_name = os.path.split(uri)
            original_file_size = file_statistics.st_size // 2 ** 10
            original_access_time = cls.timestamp_to_iso8601(
                file_statistics.st_atime
            )

        if cls._infer_protocol(uri) is Protocols.HTTP:
            parsed_url = urlparse(uri)

            file_statistics = cls._get_url(uri, info=True)

            original_path, original_file_name = os.path.split(
                f'{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}'
            )

            try:
                measured_size = int(
                    file_statistics.get_all(
                        'Content-Length'
                    )[0]
                )
            except TypeError:
                measured_size = len(cls._get_url(uri))

            original_file_size = measured_size // (2 ** 10)

            try:
                original_access_time = file_statistics.get_all('Date')[0]
            except TypeError:
                original_access_time = datetime.strftime(datetime.now(), '%c')

        return {
            'original_file_name': original_file_name,
            'original_path': original_path,
            'original_file_size': f'{original_file_size}K',
            'original_access_time': original_access_time
        }

    @staticmethod
    def timestamp_to_iso8601(timestamp):
        """Converts an Unix epoch integer to ISO8601 format.
        The converted date is in UTC (GMT-0).

        Parameters
        ----------
        timestamp: int
            With a integer timestamp (seconds after the zero hour of 1970).

        Returns
        -------
        str:
            A string with formatted date using the mask %Y-%m-%dT%H:%M:%S
        """
        return datetime.utcfromtimestamp(timestamp).\
            strftime('%Y-%m-%dT%H:%M:%S')

    @staticmethod
    def _infer_protocol(uri):
        """Discovers the protocol which the passed uri may pertain.

        Parameters
        ----------
        uri: str
            URI that describes the file location.

        Returns
        -------
        Enum:
            Referencing the discovered protocol
        """
        if urlparse(uri).scheme.startswith('http'):
            return Protocols.HTTP

        if urlparse(uri).scheme == 'file' or os.path.exists(uri):
            return Protocols.FILE

        return Protocols.UNKNOWN

    def infer_file_type_from_uri(self, uri, mimetype=False):
        """Infers the file type from an uri, with optional mime type discovery.

        Parameters
        ----------
        uri: str
            With the file location to be analyzed.

        mimetype: bool
            If should be returned also the discovered mime type.

        Returns
        -------
        str or tuple:
            str if mimetype is False, this case describing the format
            or tuple if mimetype is True, adding the mimetype to the return.

        Raises
        ------
        FileTypeNotSupportedYet
            For a not supported file type.

        Example
        -------
        ``infer_file_type_from_uri('image.jpg') # Returns 'JPG'``

        ``infer_file_type_from_uri('image.jpg, mimetype=True') # Returns ('JPG'
        , 'image/jpeg')``
        """
        guessed_init_type = mimetypes.guess_type(uri)

        if guessed_init_type[0] == 'application/x-tar':
            return self.tar_compressed_types_resolve(
                guessed_init_type[1],
                mimetype
            )

        file_bytes = self.get(uri)

        try:
            if file_bytes is Protocols.UNKNOWN:
                raise FileTypeNotSupportedYet(f'File {uri} not yet supported.')

            return self.guess_file_type(file_bytes, mimetype=mimetype)
        except FileTypeNotSupportedYet as file_type_not_supported_yet:
            file_type = mimetypes.guess_type(uri)[0]

            if file_type:
                return file_type if mimetype else \
                    file_type.split('/')[1].upper()

            raise file_type_not_supported_yet

    @classmethod
    def scan_csv(cls, uri):
        """Scanner for CSV text files.

        Parameters
        ----------
        uri: str
            Where CSV file resides

        Yields
        ------
        str:
            With the discovery file paths

        Raises
        ------
        FileTypeNotSupportedYet
            For a not supported file type.
        """
        csv_string = cls.get(uri)

        if csv_string is Protocols.UNKNOWN:
            raise FileTypeNotSupportedYet(f'File {uri} not yet supported.')
        csv_string = csv_string.decode('utf-8').splitlines()

        reader = csv.reader(csv_string)

        for row in reader:
            yield row[0]

    @classmethod
    def scan_csv_gzip(cls, uri):
        """Scanner for CSV formatted text files, compressed with
        gzip algorithm.

        Parameters
        ----------
        uri: str
            Where csv file resides

        Yields
        -------
        str:
            With the discovery file paths
        """
        file_bytes = cls.get(uri)

        gzip_file = gzip.decompress(
            file_bytes
        ).decode('utf-8').splitlines()

        for row in gzip_file:
            yield row.replace('\n', '')

    @staticmethod
    def safe_temp_file(**kwargs):
        """Creates a secure temporary file name, which means a file with an
        unique name.

        If a file with the same name is found, it's deleted before generating
        a new unique name.

        Parameters
        ----------
        file_name: str
            Defining a temporary file to assert.

        Returns
        -------
        str:
            With the complete path of the new temporary file created.
        """
        temp_dir = tempfile.gettempdir()

        if kwargs.get('file_name'):
            file_name = os.path.join(temp_dir, kwargs.get('file_name'))
        else:
            file_name = os.path.join(temp_dir, str(uuid.uuid4()))

        if os.path.exists(file_name):
            os.remove(file_name)

        return file_name

    @classmethod
    def scan_csv_bzip2(cls, uri):
        """Scanner for CSV text files, compressed with Bzip2 algorithm.

        Parameters
        ----------
        uri: str
            Where the bzip2 csv file resides.

        Yields
        -------
        str:
            With the discovered file paths.
        """
        file_bytes = cls.get(uri)

        bz2_file = bz2.decompress(
            file_bytes
        ).decode('utf-8').splitlines()

        for row in bz2_file:
            yield row.replace('\n', '')

    @classmethod
    def scan_csv_zip(cls, uri):
        """Scanner for CSV text files, compressed with Zip algorithm.

        Parameters
        ----------
        uri: str
            Where zipped csv file resides.

        Yields
        -------
        str:
            With the discovered file paths.
        """
        file_bytes = BytesIO(cls.get(uri))

        with zipfile.ZipFile(file_bytes) as ffile:
            for listed_file in ffile.namelist():
                with zipfile.ZipFile(file_bytes).open(listed_file) as ufile:
                    for row in ufile:
                        yield row.decode('utf-8').replace('\n', '')

    @classmethod
    def scan_csv_xz(cls, uri):
        """Scanner for CSV text files, compressed with Lzma algorithm.

        Parameters
        ----------
        uri: str
            Where csv xz file resides.

        Yields
        -------
        str:
            With the discovered file paths.
        """
        file_bytes = cls.get(uri)

        xz_file = lzma.decompress(
            file_bytes
        ).decode('utf-8').splitlines()

        for row in xz_file:
            yield row.replace('\n', '')

    def scan(self, uri):
        """Returns a validated uri, resolving several cases related
        to file types and methods for reading it. It also choose the best
        discovery method.

        Parameters
        ----------
        uri: str
            A file or directory to scan.

        Yields
        -------
        str:
            With actual underlying data like bytes internally on the compressed
            file container.
        """
        tar_compressed_file_readers = {
            'TZ2': 'r{stream_type}bz2',
            'TGZ': 'r{stream_type}gz',
            'TXZ': 'r{stream_type}xz'
        }

        type_to_scanner = {
            'CSV': self.scan_csv,
            'PLAIN': self.scan_csv,
            'GZP': self.scan_csv_gzip,
            'ZIP': self.scan_csv_zip,
            'BZ2': self.scan_csv_bzip2,
            'LXZ': self.scan_csv_xz,
            'TXZ': self.scan_compressed_tar_file,
            'TZ2': self.scan_compressed_tar_file,
            'TGZ': self.scan_compressed_tar_file
        }

        try:
            inferred_type = self.infer_file_type_from_uri(uri)
            scanner_function = type_to_scanner[inferred_type]

            if inferred_type in ('TXZ', 'TZ2', 'TGZ'):
                yield from scanner_function(
                    uri, tar_compressed_file_readers[inferred_type]
                )
            else:
                yield from scanner_function(uri)
        except KeyError as key_error:
            raise FileScanNotPossible(f'{uri} scan is impossible.') \
                from key_error
        except (IsADirectoryError, PermissionError):
            for root, _, files in os.walk(uri):
                for ffile in files:
                    yield os.path.join(root, ffile)

    @classmethod
    def scan_compressed_tar_file(cls, uri, file_reader):
        """Scans a compressed tar file.

        Parameters
        ----------
        uri: str
            Location where the tar file is stored.

        file_reader: str
            Suitable file reader type.

        Yields
        ------
        str:
            Paths of the already untarred files on the temporary directory.
        """
        # Won't explicitly remove the temporary directory
        temp_dir = tempfile.mkdtemp()

        inferred_protocol = cls._infer_protocol(uri)

        if inferred_protocol is Protocols.FILE:
            with tarfile.open(
                uri,
                file_reader.format(stream_type=':')
            ) as tar_file:
                tar_file.extractall(temp_dir)

        elif inferred_protocol is Protocols.HTTP:

            with urlopen(uri) as opened_url, tarfile.open(
                fileobj=opened_url,
                mode=file_reader.format(stream_type='|')
            ) as tar_file:
                for member in cls.progress(tar_file):
                    tar_file.extract(member, path=temp_dir)

        for root, _, files in os.walk(temp_dir):
            for ffile in files:
                yield os.path.join(root, ffile)

    @staticmethod
    def _get_terminal_size():
        """Returns the number of columns of current terminal.

        Returns
        -------
        int:
            Cointaning the number of columns on the current terminal emulator.
        """
        try:
            terminal_half_columns = os.get_terminal_size().columns
        except OSError:
            terminal_half_columns = 128

        return terminal_half_columns

    @classmethod
    def progress(cls, iterable, precise=False, message=None):
        """Utility method to interface process progress bar with users.
        It supports two way of unpacking the iterable, throughout ``precise``
        parameters. If ``precise`` is set to ``False`` (which is the default),
        the parameter ``iterable`` will be unpacked as is. This leads to an
        imprecise rolling of items (in other words, the method doesn't know
        apriori the total number of elements in ``iterable``). Otherwise, if
        ``precise`` is set to ``True``, an ``iterable`` which is not unpacked
        (like a ``generator``) will be first unrolled, which is much slower in
        some cases, but leads to a precise progress bar.

        Parameters
        ----------
        iterable: iter
            Objects which supports iteration.

        precise: bool
            If the progress should be precise
            (with actual percentage of completion) or just an
            interface during process running.

        message: str
            A custom message when reporting progress.

        Yields
        ------
        type:
            It returns any type on the iterable passed through the ``iterable``
            parameter.
        """
        clean_size = 0

        if not message:
            message = 'Processing, please wait:'

        if precise:
            try:
                count = len(iterable)
            except TypeError:
                iterable = [*iterable]
                count = len(iterable)

            print(
                termcolor.colored(
                    message,
                    color='green',
                    attrs=['bold']
                )
            )

            for index, value in enumerate(iterable):

                percentage = (index + 1) / count

                current_terminal_size = cls._get_terminal_size()

                clean_size = max(clean_size, current_terminal_size)

                columns_to_draw = percentage << intmul >> \
                    current_terminal_size // 4

                bar_to_draw = '🟦' * columns_to_draw
                percent_message = f'{percentage << intmul >> 100}%'

                print(f"\033[A{' ' * clean_size}\033[A")
                print(message, bar_to_draw, percent_message)

                yield value
        else:
            clocks = '🕛🕐🕑🕒🕓🕔🕕🕖🕗🕘🕙🕚'

            for index, value_clock in enumerate(
                    zip(iterable, cycle(clocks))
            ):
                item_message = message + f' {index + 1} items.'

                clean_size = max(clean_size, cls._get_terminal_size())

                print(f"\033[A{' ' * clean_size}\033[A")
                print(value_clock[1], item_message)

                yield value_clock[0]

    @staticmethod
    def resolve_path_end(path):
        """Removes directory separators from the end of
        some path (if it exists).

        Parameters
        ----------
        path: str
            Complete path to be analyzed.

        Returns
        -------
        str:
            A path without an ending character.
        """
        if path[-1] == os.path.sep:
            path = list(path)
            _ = path.pop()

            return ''.join(path)

        return path

    def dump(self, data_dir, output_dir):
        """Reads an entire database tree, compress and export it.

        Parameters
        ----------
        data_dir: str
            The directory containing all database assets.

        output_dir: str
            Location where to save the exported dump file.
        """
        data_dir = self.resolve_path_end(data_dir)
        base_name = os.path.basename(data_dir)
        dump_name = '.'.join((base_name, 'pupyl'))
        dump_path = os.path.join(output_dir, dump_name)

        with tarfile.open(dump_path, 'w:xz') as tar:
            for path in self.progress(self.scan(data_dir)):
                path_split = path.split(os.path.sep)
                archive_name = os.path.join(
                    *path_split[path_split.index(base_name):]
                )

                tar.add(path, arcname=archive_name)

    @staticmethod
    def bind(dump_file, output_dir):
        """Reads a packaged database and import it.

        Parameters
        ----------
        dump_file: str
            The directory containing all database assets.

        output_dir: str
            Location where to save the export file.
        """
        with tarfile.open(dump_file) as dfile:
            dfile.extractall(output_dir)

    @staticmethod
    def extension(uri):
        """Extract extension from ``uri``

        Parameters
        ----------
        uri: str
            URI to extract the file extension.
        """
        return os.path.splitext(uri)[-1]



"""Unit tests related to duplex.file_io module."""
import os
import csv
import tempfile
import mimetypes
from os import walk
from pathlib import Path
from unittest import TestCase
from tarfile import is_tarfile
from urllib.error import URLError
from os.path import abspath, exists, join

from pupyl.duplex.file_io import FileIO, Protocols
from pupyl.duplex.file_types import TarCompressedTypes
from pupyl.duplex.exceptions import FileTypeNotSupportedYet, \
    FileScanNotPossible


TEST_DIR = 'tests/'
TEST_SCAN_DIR = TEST_DIR + 'test_scan/'
TEST_UNKNOWN = 'unk://path'
TEST_UNSUPPORTED_FILE_TYPE = abspath(f'{TEST_DIR}not_image.inv')
TEST_LOCAL = abspath(f'{TEST_DIR}test_image.jpg')
TEST_URL = 'https://upload.wikimedia.org/wikipedia/commons/' + \
    'thumb/e/e4/Cheshm-Nazar.JPG/320px-Cheshm-Nazar.JPG'
TEST_URL_NO_DATE = 'http://images.protopage.com/view/572714/' + \
    'axuvb8oxm7liskynxggfczfus.jpg'
TEST_URL_FORBIDDEN = 'https://cutt.ly/hWtd8dN'
TEST_URL_TIMEOUT = 'http://www.pedigree.com.sg/breeds/images/' + \
    'norwich_terr_02.jpg'
TEST_CSV = abspath(TEST_SCAN_DIR + 'test_csv.csv')
TEST_CSV_ZIP = abspath(TEST_SCAN_DIR + 'test_csv.csv.zip')
TEST_CSV_GZ = abspath(TEST_SCAN_DIR + 'test_csv.csv.gz')
TEST_CSV_BZ2 = abspath(TEST_SCAN_DIR + 'test_csv.csv.bz2')
TEST_CSV_XZ = abspath(TEST_SCAN_DIR + 'test_csv.csv.xz')
TEST_TAR_LOCATION = TEST_DIR + 'tar_files'


def util_test_csv(path):
    """
    Closure to mimic scan_csv behaviour.
    """
    with open(path) as ffile:
        reader = csv.reader(ffile)

        for row in reader:
            yield row[0]


class TestCases(TestCase):
    """Unit tests over special cases."""

    def test__get_local_unsuccessful(self):
        """Unit test for _get_local method, local unsuccessful case."""
        with self.assertRaises(IOError):
            FileIO._get_local(TEST_UNKNOWN)

    def test_infer_file_type_from_uri_unknown(self):
        """Unit test infer_file_type_from_uri, unknown case."""
        with self.assertRaises(FileTypeNotSupportedYet):
            file_io = FileIO()

            file_io.infer_file_type_from_uri(TEST_UNKNOWN)

    def test_scan_unknown_file(self):
        """Unit test for scan method, unknown file case."""
        file_io = FileIO()

        with self.assertRaises(FileTypeNotSupportedYet):
            next(file_io.scan(TEST_UNKNOWN))

    def test_scan_file(self):
        """Unit test for scan method, file case."""
        file_io = FileIO()

        with self.assertRaises(FileScanNotPossible):
            assert next(file_io.scan(TEST_LOCAL)) == TEST_LOCAL

    def test_scan_csv_unknown_file(self):
        """Unit test for scan_csv method, unknown file case."""
        file_io = FileIO()

        with self.assertRaises(FileTypeNotSupportedYet):
            next(file_io.scan_csv(TEST_UNKNOWN))

    def test_infer_file_type_from_uri_unsupported(self):
        """
        Unit test for method infer_file_type_from_uri,
        unsupported file type case.
        """
        file_io = FileIO()

        with self.assertRaises(FileTypeNotSupportedYet):
            file_io.infer_file_type_from_uri(
                TEST_UNSUPPORTED_FILE_TYPE,
                mimetype=True
            )

    def test__get_url_timeout(self):
        """Unit tests for _get_url method, timeout case."""
        with self.assertRaises(URLError):
            FileIO._get_url(TEST_URL_TIMEOUT)


def test_safe_temp_file():
    """Unit test for method safe_temp_file."""
    test_temp_file_name = FileIO.safe_temp_file()

    assert not exists(test_temp_file_name)


def test_safe_temp_file_exists():
    """Unit test for method safe_temp_file, file exists case."""
    test_temp_file_name = 'just_a_temp_file.txt'

    Path(join(tempfile.gettempdir(), test_temp_file_name)).touch()

    _ = FileIO.safe_temp_file(file_name=test_temp_file_name)

    assert not exists(test_temp_file_name)


def test__infer_protocol_http():
    """Unit tests for _infer_protocol method, http case."""
    assert FileIO._infer_protocol(TEST_URL) is Protocols.HTTP


def test__infer_protocol_local():
    """Unit tests for _infer_protocol method, local case."""
    assert FileIO._infer_protocol(TEST_LOCAL) is Protocols.FILE


def test__infer_protocol_unknown():
    """Unit tests for _infer_protocol method, unknown case."""
    assert FileIO._infer_protocol(TEST_UNKNOWN) is Protocols.UNKNOWN


def test__get_url_successful():
    """Unit tests for _get_url method, successful case."""
    assert isinstance(FileIO._get_url(TEST_URL), bytes)


def test__get_url_unsuccessful():
    """Unit tests for _get_url method, unsuccessful case."""
    try:
        FileIO._get_url(TEST_UNKNOWN)
    except IOError:
        assert True


def test__get_url_large_file():
    """Unit test for method _get_url, large file case."""
    test_url_big_file = 'https://speed.hetzner.de/10GB.bin'

    assert FileIO._get_url(
        test_url_big_file,
        headers={'User-Agent': 'Mozilla/5.0'}
    ) == b''


def test__get_local_big_file():
    """Unit test for method _get_local, big file case."""
    with tempfile.NamedTemporaryFile() as temp_file:
        while int(
            os.path.getsize(temp_file.name) / (2 ** 30)
        ) < FileIO.max_file_size:
            temp_file.write(b'a' * (2 ** 16))

        assert FileIO._get_local(temp_file.name) == b''


def test_get_http():
    """
    Unit test for get method, http case
    """
    assert isinstance(FileIO.get(TEST_URL), bytes)


def test_get_local():
    """
    Unit test for get method, http case
    """
    assert isinstance(FileIO.get(TEST_LOCAL), bytes)


def test__get_local_successful():
    """
    Unit test for _get_local method, local successful case
    """
    assert isinstance(FileIO._get_local(TEST_LOCAL), bytes)


def test_get_unknown():
    """
    Unit test for get method, unknown case
    """
    assert FileIO.get(TEST_UNKNOWN) is Protocols.UNKNOWN


def test_get_file_scheme():
    """Unit test for get method, file scheme case."""
    test_file = f'file:///{TEST_LOCAL}'

    assert isinstance(FileIO.get(test_file), bytes)


def test_scan_directory():
    """
    Unit test for scan method, directory case
    """
    file_io = FileIO()

    test_against_tree = [
        abspath(f'{TEST_SCAN_DIR}{ffile}')
        for ffile in [*walk(TEST_SCAN_DIR)][0][-1]
    ]

    test_current_tree = [*file_io.scan(abspath(TEST_SCAN_DIR))]

    assert test_current_tree == test_against_tree


def test_scan_csv():
    """ Unit test for scan method, txt or csv case. """

    file_io = FileIO()

    for method_row, util_row in zip(file_io.scan(TEST_CSV),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_scan_scan_gzip():
    """ Unit test for scan method, gzip case. """
    file_io = FileIO()

    for method_row, util_row in zip(file_io.scan(TEST_CSV_GZ),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_scan_scan_zip():
    """ Unit test for scan method, zip case. """
    file_io = FileIO()

    for method_row, util_row in zip(file_io.scan(TEST_CSV_ZIP),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_scan_scan_bzip2():
    """ Unit test for scan method, gzip case. """
    file_io = FileIO()

    for method_row, util_row in zip(file_io.scan(TEST_CSV_BZ2),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_scan_scan_xz():
    """ Unit test for scan method, lzma case. """
    file_io = FileIO()

    for method_row, util_row in zip(file_io.scan(TEST_CSV_XZ),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_scan_csv_gzip():
    """ Unit test for scan_csv_gzip method. """
    for method_row, util_row in zip(FileIO.scan_csv_gzip(TEST_CSV_GZ),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_scan_csv_xz():
    """ Unit test for scan_csv_xz method. """
    for method_row, util_row in zip(FileIO.scan_csv_xz(TEST_CSV_XZ),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_scan_csv_zip():
    """ Unit test for scan_csv_zip method. """
    for method_row, util_row in zip(FileIO.scan_csv_zip(TEST_CSV_ZIP),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_scan_csv_bzip2():
    """ Unit test for scan_csv_bzip2 method. """
    for method_row, util_row in zip(FileIO.scan_csv_bzip2(TEST_CSV_BZ2),
                                    util_test_csv(TEST_CSV)):
        assert method_row == util_row


def test_get_metadata_http():
    """Unit test for method get_metadata, http case."""
    test_metadata = {
        'original_file_name': '320px-Cheshm-Nazar.JPG',
        'original_path': """https://upload.wikimedia.org/wikipedia/commons/
        thumb/e/e4/Cheshm-Nazar.JPG""".replace('\n        ', ''),
        'original_file_size': '9K'
    }

    test_request_metadata = FileIO.get_metadata(TEST_URL)

    del test_request_metadata['original_access_time']

    assert test_metadata == test_request_metadata


def test_get_metadata_http_no_date():
    """Unit test for method get_metadata, http and not date case."""
    test_metadata = {
        'original_file_name': 'axuvb8oxm7liskynxggfczfus.jpg',
        'original_path': """http://images.protopage.com/view/
        572714""".replace('\n        ', '')
    }

    test_request_metadata = FileIO.get_metadata(TEST_URL_NO_DATE)

    del test_request_metadata['original_access_time']
    del test_request_metadata['original_file_size']

    assert test_metadata == test_request_metadata


def test_get_metadata_local():
    """Unit test for method get_metadata, local case."""
    test_metadata = {
        'original_file_name': 'test_image.jpg',
        'original_path': abspath('tests'),
        'original_file_size': '5K'
    }

    test_local_metadata = FileIO.get_metadata(TEST_LOCAL)

    del test_local_metadata['original_access_time']

    assert test_metadata == test_local_metadata


def test_timestamp_to_iso8601():
    """Unit test for method timestamp_to_iso8601"""
    test_timestamp = 1591123280
    expected_return = '2020-06-02T18:41:20'

    assert FileIO.timestamp_to_iso8601(test_timestamp) == expected_return


def test_infer_file_type_from_uri_with_mimetype():
    """
    Unit test for method infer_file_type_from_uri,
    with mimetype case
    """
    file_io = FileIO()

    _, mime = file_io.infer_file_type_from_uri(
        TEST_LOCAL,
        mimetype=True
    )

    assert mime == 'image/jpeg'


def test_infer_file_type_from_uri_no_mimetype():
    """
    Unit test for method infer_file_type_from_uri,
    no mimetype case
    """
    file_io = FileIO()

    assert file_io.infer_file_type_from_uri(
        TEST_LOCAL,
        mimetype=False
    ) == 'JPG'


def test_infer_file_type_from_uri_remote():
    """
    Unit test for method infer_file_type_from_uri,
    remote case.
    """
    file_io = FileIO()

    assert file_io.infer_file_type_from_uri(TEST_URL) == 'JPG'


def test_infer_file_type_tar_files():
    """Unit test for method infer_file_type_from_uri, tar file case."""
    file_io = FileIO()

    for ffile in os.listdir(TEST_TAR_LOCATION):
        test_file_type = mimetypes.guess_type(
            os.path.join(
                TEST_TAR_LOCATION,
                ffile
            )
        )[1]

        assert file_io.infer_file_type_from_uri(
            os.path.join(
                TEST_TAR_LOCATION,
                ffile
            ),
            mimetype=True
        ) == TarCompressedTypes.mime(test_file_type)

        assert file_io.infer_file_type_from_uri(
            os.path.join(
                TEST_TAR_LOCATION,
                ffile
            ),
            mimetype=False
        ) == TarCompressedTypes.name(test_file_type)


def test_progress_not_precise():
    """Unit test for method progress, not precise case."""
    test_generator = range(10)
    test_unpacked = [*test_generator]

    test_result_generator = FileIO.progress(test_generator, precise=False)
    test_result_unpacked = FileIO.progress(test_unpacked, precise=False)

    for t_gen, r_gen in zip(test_generator, test_result_generator):
        assert t_gen == r_gen

    for t_unp, r_unp in zip(test_unpacked, test_result_unpacked):
        assert t_unp == r_unp


def test_progress_precise():
    """Unit test for method progress, not precise case."""

    def test_gen():
        """Closure to test functions which returns generators."""
        for value in range(10):
            yield value

    test_generator = range(10)
    test_unpacked = [*test_generator]

    test_result_generator = FileIO.progress(test_generator, precise=True)
    test_result_unpacked = FileIO.progress(test_unpacked, precise=True)
    test_result_func_gen = FileIO.progress(test_gen(), precise=True)

    for t_gen, r_gen in zip(test_generator, test_result_generator):
        assert t_gen == r_gen

    for t_unp, r_unp in zip(test_unpacked, test_result_unpacked):
        assert t_unp == r_unp

    for t_fgen, r_fgen in zip(test_generator, test_result_func_gen):
        assert t_fgen == r_fgen


def test_resolve_path_end():
    """Unit test for method test_resolve_path_end."""
    test_path_with_sep = '/just/a/test/path/'
    test_path_without_sep = '/just/a/test/path'

    expected_return = '/just/a/test/path'

    assert FileIO.resolve_path_end(test_path_with_sep) == expected_return
    assert FileIO.resolve_path_end(test_path_without_sep) == expected_return


def test_dump():
    """Unit test for method dump."""
    file_io = FileIO()

    file_io.dump(TEST_SCAN_DIR, tempfile.gettempdir())

    assert is_tarfile(
        os.path.join(
            tempfile.gettempdir(),
            '.'.join(
                (
                    os.path.basename(
                        file_io.resolve_path_end(TEST_SCAN_DIR)
                    ),
                    'pupyl'
                )
            )
        )
    )


def test_bind():
    """Unit test for method bind."""
    file_io = FileIO()

    file_io.dump(TEST_SCAN_DIR, tempfile.gettempdir())

    file_io.bind(
        os.path.join(
            tempfile.gettempdir(),
            '.'.join(
                (
                    os.path.basename(
                        file_io.resolve_path_end(TEST_SCAN_DIR)
                    ),
                    'pupyl'
                )
            )
        ),
        os.path.join(tempfile.gettempdir(), TEST_SCAN_DIR)
    )

    assert os.path.isdir(
        os.path.join(
            tempfile.gettempdir(),
            TEST_SCAN_DIR
        )
    )


def test_scan_tar_files():
    """Unit test for method scan, compressed tar file case."""
    file_io = FileIO()

    for ffile in os.listdir(TEST_TAR_LOCATION):
        for dfile in file_io.scan(
            os.path.join(TEST_TAR_LOCATION, ffile)
        ):
            assert os.path.exists(dfile)


def test_scan_compressed_tar_file_local():
    """Unit test for method scan_compressed_tar_file, local case."""
    test_tar_compressed_file_readers = {
        'TZ2': 'r:bz2',
        'TGZ': 'r:gz',
        'TXZ': 'r:xz'
    }

    file_io = FileIO()

    for ffile in os.listdir(TEST_TAR_LOCATION):
        test_file_type = mimetypes.guess_type(
            os.path.join(
                TEST_TAR_LOCATION,
                ffile
            )
        )[1]

        for dfile in file_io.scan_compressed_tar_file(
            os.path.join(TEST_TAR_LOCATION, ffile),
            test_tar_compressed_file_readers[
                TarCompressedTypes.name(test_file_type)
            ]
        ):
            assert os.path.exists(dfile)


def test_scan_compressed_tar_file_http():
    """Unit test for method scan_compressed_tar_file, http case."""
    test_uri = 'http://localhost:8888/images.tar.xz'
    test_file_reader = 'r|xz'

    file_io = FileIO()

    for ffile in file_io.scan_compressed_tar_file(test_uri, test_file_reader):
        assert os.path.exists(ffile)


def test_request_http_not_found():
    """Unit test for method _get_url, http 404 case."""
    test_uri = 'http://localhost:8888/notfound'

    assert FileIO._get_url(test_uri) == b''


def test__file_scheme_to_path():
    """Unit test for method _file_scheme_to_path."""
    test_uri = 'file:///path/to/a/test/file'

    assert FileIO._file_scheme_to_path(test_uri) == '/path/to/a/test/file'


def test__get_url_forbidden():
    """Unit test for method _get_url, HTTP 403 case."""
    assert isinstance(FileIO._get_url(TEST_URL_FORBIDDEN), bytes)



"""Operations and storage for images."""

import os
import json
import warnings
import termcolor
from shutil import copy, move

from pupyl.duplex.image import ImageIO


class ImageDatabase(ImageIO):
    """Handling storage and database operations for images."""

    def __init__(self, import_images, data_dir=None, **kwargs):
        """Image storage and operations.

        Parameters
        ----------
        import_images: bool
            If images must be imported (copied) to the internal database or
            not.

        data_dir: str
            Location to save the image storage files and assets. If a value is
            ommited for this parameter, will be created a new temporary folder
            in the underlying (operating system) default temporary directory.

        bucket_size: int
            Defines the number of files per bucket inside the database. Since
            each file and associated assets are saved together, splitting up
            the directories will help avoid issues like ``Too many files``,
            also allowing read parallelization of assets among others features.
            In other words, this parameter describes how many image files will
            be saved on an internal database directory before starting to save
            to another new one.

        image_size: tuple
            Defines the dimensions (in pixels, width x height)
            of saved images on the database. Only has some effect if
            ``import_images`` is True. Case a resize happens, the aspect ratio
            of the original image will be preserved, hence ``image_size`` is an
            approximation. In other words, the image will be resized to
            dimensions close to ``800x600``, but using one pair of dimensions
            that not offends the image aspect.

        Caution
        -------
        If no value is passed to ``data_dir``, all database assets will be
        created on the defined temporary directory. By doing this, be advised
        that all your image search will (probably) vanish after a system
        reboot. If you don't want that this happens, please, define a
        non-volatile ``data_dir``.
        """
        self._import_images = import_images

        if data_dir:
            self._data_dir = os.path.normpath(data_dir)
        else:
            self._data_dir = self.safe_temp_file()

        if kwargs.get('bucket_size'):
            self._bucket_size = kwargs['bucket_size']
        else:
            self._bucket_size = 1000

        if self._import_images:
            if kwargs.get('image_size'):
                self._image_size = kwargs['image_size']
            else:
                self._image_size = (800, 600)

        os.makedirs(self._data_dir, exist_ok=True)

    def __getitem__(self, position):
        """Returns the item at index.

        Parameters
        ----------
        position: int
            The position inside database to return.

        Returns
        -------
        dict:
            With some metadata related to the item.

        Example
        -------
        ``img_db = ImageDatabase(import_images=True, data_dir='pupyl')``

        ``img_db[10]``

        ``# May return:``

        ``{'original_file_name': '2610447919_1b91946bd1.jpg',``
        ``'original_path': '/tmp/tmpekd0cuie',``
        ``'original_file_size': '52K',``
        ``'original_access_time': '2021-06-14T19:07:27',``
        ``'id': 10}``
        """
        return self.load_image_metadata(position)

    def __len__(self):
        """Return how many items are indexed in the database.

        Returns
        -------
        int:
            Describing how may images are indexed on the current database.

        Example
        -------
        ``img_db = ImageDatabase(import_images=True, data_dir='pupyl')``

        ``len(img_db) # May return 709``
        """
        return len([*self.list_images()])

    @property
    def import_images(self):
        """Getter for import_images property.

        Returns
        -------
        bool:
            If images should be imported into the current database or not.
        """
        return self._import_images

    @import_images.setter
    def import_images(self, import_it):
        """Setter for import_images property.

        Parameters
        ----------
        import_it: bool
            If images should be imported into the current database or not.
        """
        self._import_images = import_it

    @property
    def bucket_size(self):
        """Getter for bucket_size property.

        Returns
        -------
        int:
            With how many files per bucket will be stored.
        """
        return self._bucket_size

    @bucket_size.setter
    def bucket_size(self, size):
        """Setter for bucket_size property.

        Parameters
        ----------
        size: int
            Defines how many files per bucket will be saved.
        """
        self._bucket_size = size

    @property
    def image_size(self):
        """Getter for image_size property.

        Returns
        -------
        tuple:
            Describing the internal (approximated) dimensions of each image. If
            ``_import_images`` is undefined, returns by default ``(800x600)``.
        """
        if self._import_images:
            return self._image_size

        return (800, 600)

    @image_size.setter
    def image_size(self, dimensions):
        """Setter for image_size property.

        Parameters
        ----------
        dimensions: tuple
            With dimensions for a image.
        """
        if self._import_images:
            self._image_size = dimensions

    def what_bucket(self, index):
        """Discovers in what bucket the file should be saved.

        Parameters
        ----------
        index: int
            The index that references an image in the database.

        Returns
        -------
        int:
            With the bucket number that the image is saved.
        """
        return index // self._bucket_size

    def mount_file_name(self, index, **kwargs):
        """Creates the full name path that the file will be saved inside
        database.

        Parameters
        ----------
        index: int
            The indexer id associated with the file.

        extension: str
            Describing the file extension.

        Returns
        -------
        str:
            With the full path inside the database.
        """
        if kwargs.get('extension'):
            extension = kwargs['extension'][1:] \
                if kwargs['extension'][0] == '.' else kwargs['extension']

            return os.path.join(
                self._data_dir,
                str(self.what_bucket(index)),
                f'{index}.{extension}'
            )

        return os.path.join(
            self._data_dir,
            str(self.what_bucket(index)),
            f'{index}'
        )

    def load_image_metadata(self, index, **kwargs):
        """Loads the metadata for an image inside the database.

        Parameters
        ----------
        index: int
            Regarding the position of some image inside database.

        filtered: iterable
            Describing which fields to filter (or select) for return.

        distance: float
            The distance between the tensor represented by ``index``
            and the ``query`` image.

        Returns
        -------
        dict
            Containing the parsed json file.

        Raises
        ------
        IndexError:
            When ``index`` is not found.
        """
        result_file_name = self.mount_file_name(index, extension='json')

        try:
            with open(result_file_name, encoding='utf-8') as json_file:
                metadata = json.load(json_file)

                if kwargs.get('filtered'):
                    metadata = {
                        key: value
                        for key, value in metadata.items()
                        if key in kwargs['filtered']
                    }

                if kwargs.get('distance'):
                    metadata['distance'] = kwargs['distance']

                return metadata

        except FileNotFoundError:
            raise IndexError(
                f'Index not found: {index}'
            ) from FileNotFoundError

    def save_image_metadata(self, index, uri):
        """Stores image metadata information retrieved from the file.

        Parameters
        ----------
        index: int
            The index related to the image.

        uri: str
            Location where the image is stored.
        """
        bytess = self.get(uri)
        result_file_name = self.mount_file_name(index, extension='json')

        os.makedirs(os.path.dirname(result_file_name), exist_ok=True)

        if self.is_image(bytess):
            with open(result_file_name, 'w', encoding='utf-8') as json_file:
                metadata = self.get_metadata(uri)
                metadata['id'] = index
                metadata['internal_path'] = self.mount_file_name(
                    index, extension=self.extension(uri)
                )

                json.dump(metadata, json_file)

    def insert(self, index, uri):
        """Inserts an image into the database.

        Parameters
        ----------
        index: int
            The index number attributed to the image.

        uri: str
            Where the original file is located.
        """
        if self._import_images:
            mounted_file = self.mount_file_name(
                index, extension=self.extension(uri)
            )

            if not os.path.exists(mounted_file):
                compression_not_supported = not self.os_supports_compression()
                if (
                    self.is_animated_gif(uri) or compression_not_supported
                ):
                    warnings.warn(
                        termcolor.colored(
                            f'Storing {uri} with original compression due '
                            'to underlying platform incompatibility.',
                            color='yellow',
                            attrs=['bold']
                        ),
                        RuntimeWarning
                    )
                    image = self.get_image(uri)
                else:
                    image = self.compress(
                        self.size(
                            uri, new_size=self._image_size, keep_aspect=True
                        )
                    )

                self.save_image(mounted_file, image)

        self.save_image_metadata(index, uri)

    def remove(self, index):
        """Removes the image at ``index``.

        Parameters
        ----------
        index: int
            The image index to remove from database.

        Danger
        ------
        Use this method with caution. The deleted image cannot be restored.
        No prompt are shown before deletion.

        Attention
        ---------
        Be advised that this operation is linear on the index size
        (:math:`O(n)`). It provokes changes on the current image ``index``, for
        all indexed images. For instance, if the ``index`` at 54 was deleted,
        every image with index greater than 54 will have the ``id``
        decreased by one.
        """
        for old_id in range(index + 1, len(self)):
            new_id = old_id - 1

            # Renaming images first
            image_old_id = self.load_image_metadata(
                old_id, filtered=['internal_path']
            )['internal_path']

            image_new_id = self.load_image_metadata(
                new_id, filtered=['internal_path']
            )['internal_path']

            move(image_old_id, image_new_id)

            # Now, the metadata files
            metadata_old_id = self.mount_file_name(old_id, extension='json')
            metadata_new_id = self.mount_file_name(new_id, extension='json')

            copy(metadata_old_id, metadata_new_id)

            with open(
                metadata_new_id, 'r+', encoding='utf-8'
            ) as metadata_file:
                metadata = json.load(metadata_file)

                metadata['id'] = new_id
                metadata['internal_path'] = image_new_id

                metadata_file.seek(0)
                json.dump(metadata, metadata_file)

        os.remove(self.mount_file_name(len(self), extension='json'))

    def list_images(self, return_ids=False, top=None):
        """Returns images on current database.

        Parameters
        ----------
        return_ids: bool
            If the method should also return the file ids inside database.

        top: int
            How many pictures from image database should be listed. Not setting
            this parameter (which means not referencing it or setting it
            to zero or below) will return all images in the database.

        Yields
        ------
        tuple or str:
            If ``return_ids=True``, a ``tuple`` with ``(int, str)``
            representing respectively the index and the path inside the
            database will be returned. Otherwise, if ``return_ids=False``,
            just a ``str`` with the full path will return.
        """
        if top:
            counter = 0

        for root, _, files in os.walk(self._data_dir):
            for ffile in files:
                if self.extension(ffile) in ('.jpg', '.gif'):
                    ffile = os.path.join(root, ffile)

                    if top:
                        counter += 1

                        if counter > top:
                            return

                    if return_ids:
                        yield int(
                            os.path.splitext(
                                os.path.split(ffile)[1]
                            )[0]
                        ), ffile
                    else:
                        yield ffile

    def load_image(self, index, as_tensor=False):
        """Returns the image data at a specified index.

        Parameters
        ----------
        index: int
            The location of the image inside database.

        as_tensor: bool
            How to return the image from database: as ``bytes``
            (``as_tensor=False``) or as a ``numpy.ndarray`` tensor
            (``as_tensor=True``)

        Returns
        -------
        bytes or numpy.ndarray:
            Returns image `bytes` (`as_tensor=False`) or
            `numpy.ndarray` (`as_tensor=True`), containing
            image converted to its tensor representation.
        """
        return self.get_image(self.load_image_metadata(
            index, filtered=['internal_path'])['internal_path'], as_tensor
        )



"""Unit tests related to storage.database module."""
import tempfile
from os import sep
from shutil import copy
from unittest import TestCase
from os.path import join, relpath, exists

from pupyl.storage.database import ImageDatabase


TEST_TEMP_DIRECTORY = tempfile.gettempdir()
TEST_DIRECTORY = join('tests', 'test_database')
TEST_IMAGE = join('tests', 'test_image.jpg')
TEST_INDEX = 999

TEST_METADATA = {
    'id': 0,
    'original_file_name': 'test_image.jpg',
    'original_path': 'tests',
    'original_file_size': '5K',
    'internal_path': join('tests', 'test_database', '0', '0.jpg')
}

TEST_METADATA_HTTP = {
    'id': 1,
    'original_file_name': 'test_image.jpg',
    'original_path': 'https://domain.com/tests/',
    'original_file_size': '5K'
}


def resolve_original_path(test_metadata):
    """Acknowledge the protocol of original path data."""
    if not test_metadata['original_path'].startswith('http'):
        test_metadata['original_path'] = test_metadata[
            'original_path'].split(sep)[-1]

    return test_metadata


class TestCases(TestCase):
    """Unit tests over special cases."""

    def test___get_item___not_found(self):
        """Unit test for method __get_item__, index not found case."""
        with self.assertRaises(IndexError):
            image_database = ImageDatabase(
                import_images=True,
                data_dir=TEST_DIRECTORY
            )

            _ = image_database[999]

    def test_load_image_metadata_not_found(self):
        """Unit test for method load_image_metadata, index not found case."""
        with self.assertRaises(IndexError):
            image_database = ImageDatabase(
                import_images=True,
                directory=TEST_DIRECTORY
            )

            _ = image_database.load_image_metadata(999)


def test_image_database_definition():
    """Unit test for instance definition for class ImageDatabase."""
    test_import_images = True
    test_bucket_size = 100
    test_image_size = (640, 480)

    image_database = ImageDatabase(
        import_images=test_import_images,
        data_dir=TEST_TEMP_DIRECTORY,
        bucket_size=test_bucket_size,
        image_size=test_image_size
    )

    assert isinstance(image_database, ImageDatabase)
    assert image_database.import_images == test_import_images
    assert image_database._data_dir == TEST_TEMP_DIRECTORY
    assert image_database.bucket_size == test_bucket_size
    assert image_database.image_size == test_image_size


def test___get_item__():
    """Unit test for __get_item__ method."""

    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    test_metadata = image_database[0]
    del test_metadata['original_access_time']

    test_metadata = resolve_original_path(test_metadata)

    assert test_metadata == TEST_METADATA

    test_metadata = image_database[1]
    del test_metadata['original_access_time']

    test_metadata = resolve_original_path(test_metadata)

    assert test_metadata == TEST_METADATA_HTTP


def test___len__():
    """Unit test for __len__ method."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    assert len(image_database) == 1


def test_import_images_property():
    """Unit test for property import_images."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    test_import_images = True
    image_database.import_images = test_import_images
    assert image_database.import_images

    test_import_images = not test_import_images
    image_database.import_images = test_import_images
    assert not image_database.import_images


def test_bucket_size_property():
    """Unit test for property bucket_size."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    test_bucket_size = 999
    image_database.bucket_size = test_bucket_size
    assert image_database.bucket_size == test_bucket_size


def test_image_size_property():
    """Unit test for property image_size."""
    image_database = ImageDatabase(
        data_dir=TEST_DIRECTORY,
        import_images=True
    )

    test_image_size = (320, 200)
    image_database.image_size = test_image_size
    assert image_database.image_size == test_image_size


def test_image_size_property_no_import_images():
    """Unit test for property image_size, no import images case."""
    image_database = ImageDatabase(
        data_dir=TEST_DIRECTORY,
        import_images=False
    )

    test_image_size = (800, 600)
    assert image_database.image_size == test_image_size


def test_insert_import_images():
    """Unit test for method insert, import images case."""
    image_database = ImageDatabase(
        data_dir=TEST_TEMP_DIRECTORY,
        import_images=True
    )

    image_database.insert(10, 'tests/test_image.jpg')

    assert exists(f'{TEST_TEMP_DIRECTORY}/0/10.jpg') \
        and exists(f'{TEST_TEMP_DIRECTORY}/0/10.json')


def test_insert_import_images_gif():
    """Unit test for method insert, import gif images case."""
    image_database = ImageDatabase(
        data_dir=TEST_TEMP_DIRECTORY,
        import_images=True
    )

    image_database.insert(11, 'tests/test_gif.gif')

    assert exists(f'{TEST_TEMP_DIRECTORY}/0/11.gif') \
        and exists(f'{TEST_TEMP_DIRECTORY}/0/11.json')


def test_insert_no_import_images():
    """Unit test for method insert, no import images case."""
    image_database = ImageDatabase(
        data_dir=TEST_TEMP_DIRECTORY,
        import_images=False
    )

    image_database.insert(20, 'tests/test_image.jpg')

    assert not exists(f'{TEST_TEMP_DIRECTORY}/0/20.jpg') and \
        exists(f'{TEST_TEMP_DIRECTORY}/0/20.json')


def test_insert_image_exists_target():
    """Unit test for method insert, image exists on target path."""
    with tempfile.TemporaryDirectory() as temp_dir:
        image_database = ImageDatabase(
            data_dir=temp_dir,
            import_images=True
        )

        test_gif_path = join('tests', 'test_gif.gif')

        copy(test_gif_path, join(temp_dir, '30.gif'))

        image_database.insert(30, test_gif_path)

        assert image_database.load_image_metadata(
            30, filtered=('original_file_name')
        )['original_file_name'] == 'test_gif.gif'


def test_load_image():
    """Unit test for method load_image."""
    test_index = 0

    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    def inst_load_image(index):
        """Closure for method load_image."""
        return image_database.get_image(
            image_database.mount_file_name(index, extension='jpg')
        )

    test_image = image_database.load_image(test_index)

    assert test_image == inst_load_image(test_index)


def test_what_bucket():
    """Unit test for method what_bucket."""
    test_bucket_size = 10 ** 4

    image_database = ImageDatabase(
        import_images=True,
        bucket_size=test_bucket_size
    )

    assert image_database.what_bucket(TEST_INDEX) == \
        TEST_INDEX // test_bucket_size


def test_mount_file_name():
    """Unit test for method mount_file_name."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    expected_path = join(
        TEST_DIRECTORY,
        str(image_database.what_bucket(TEST_INDEX)),
        f'{TEST_INDEX}.json'
    )

    assert image_database.mount_file_name(
        TEST_INDEX, extension='json'
    ) == expected_path


def test_mount_file_name_no_extension():
    """Unit test for method mount_file_name, no extension case."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    expected_path = join(
        TEST_DIRECTORY,
        str(image_database.what_bucket(TEST_INDEX)),
        f'{TEST_INDEX}'
    )

    assert image_database.mount_file_name(TEST_INDEX) == expected_path


def test_load_image_metadata():
    """Unit test for method load_image_metadata."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    test_metadata = image_database.load_image_metadata(0)

    del test_metadata['original_access_time']

    test_metadata = resolve_original_path(test_metadata)

    assert test_metadata == TEST_METADATA

    test_metadata = image_database.load_image_metadata(1)

    del test_metadata['original_access_time']

    test_metadata = resolve_original_path(test_metadata)

    assert test_metadata == TEST_METADATA_HTTP


def test_load_image_metadata_filtered():
    """Unit test for method load_image_metadata, filtered input case."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    test_filter = ('id', 'original_file_size')

    test_metadata = image_database.load_image_metadata(0, filtered=test_filter)

    for key in test_metadata:
        assert key in test_filter


def test_save_image_metadata():
    """Unit test for method save_image_metadata."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    image_database.save_image_metadata(0, TEST_IMAGE)

    test_metadata = image_database.load_image_metadata(0)

    del test_metadata['original_access_time']

    test_metadata['original_path'] = relpath(test_metadata['original_path'])
    test_metadata['internal_path'] = relpath(test_metadata['internal_path'])

    assert test_metadata == TEST_METADATA


def test_load_image_metadata_distances():
    """Unit test for method load_image_metadata, returning distances case."""
    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    test_distance = .9999
    test_metadata = image_database.load_image_metadata(
        0,
        distance=test_distance
    )
    del test_metadata['original_access_time']

    local_test_metadata = TEST_METADATA
    local_test_metadata['distance'] = test_distance

    assert test_metadata == local_test_metadata


def test_list_images():
    """Unit test for method list_images."""
    expected_result = 'tests/test_database/0/0.jpg'

    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    actual_result = [*image_database.list_images()][0]

    assert expected_result == actual_result


def test_list_images_return_ids():
    """Unit test for method list_images, return ids case."""
    expected_path = 'tests/test_database/0/0.jpg'
    expected_index = 0
    expected_result = (expected_index, expected_path)

    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    actual_index_path = [
        *image_database.list_images(
            return_ids=True
        )
    ][0]

    assert actual_index_path == expected_result


def test_list_images_length_filtered():
    """Unit test for method list_images, length filter case."""
    expected_length = 1

    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    actual_result = [*image_database.list_images(top=expected_length)]

    assert len(actual_result) == expected_length


def test_list_images_less_than_count():
    """Unit test for method list_images, less than counter case."""
    expected_length = -1

    image_database = ImageDatabase(
        import_images=True,
        data_dir=TEST_DIRECTORY
    )

    actual_result = [*image_database.list_images(top=expected_length)]

    assert len(actual_result) == 0



"""Exceptions regarding embeddings module operations"""


class Error(Exception):
    """Base exception class"""


class UnknownCharacteristics(Error):
    """Raised when a not known network characteristic is passed"""


class UnknownCharacteristicsValue(Error):
    """Raised when a not known value for a characteristic is passed"""


class UnknownCharacteristicsName(Error):
    """Raised when a not known name for a characteristic is passed"""



"""Performs multiple operations over images, like resizing,loading and so on"""

from base64 import b64encode
from platform import system

from pupyl.verbosity import quiet_tf
quiet_tf()

import tensorflow
from tensorflow import io as io_ops
from tensorflow import image as image_ops

from pupyl.duplex.file_io import FileIO
from pupyl.duplex.exceptions import FileIsNotImage


class ImageIO(FileIO):
    """Operations of read and write over images."""

    @staticmethod
    def encoded_to_compressed_tensor(encoded):
        """Transforms a binary encoded string image to its compressed tensor
        form.

        Parameters
        ----------
        encoded: bytes
            The binary representation of the image.

        Returns
        -------
        numpy.ndarray
            The actual compressed image tensor.
        """
        return io_ops.decode_raw(encoded, 'uint8')

    @staticmethod
    def encoded_to_tensor(encoded):
        """Transforms a binary encoded string image to a tensor.

        Parameters
        ----------
        encoded: bytes
            The binary representation of the image.

        Returns
        -------
        numpy.ndarray
            Describing the compressed image tensor.
        """
        return io_ops.decode_image(encoded)

    @classmethod
    def get_image(cls, uri, as_tensor=False):
        """Loads an image from a specified location.

        Parameters
        ----------
        uri: str
            Location where the image are stored.

        as_tensor: bool
            If the image should be converted to its tensor representation.
            Defaults to False, which returns images to a byte representation

        Returns
        -------
        bytes or numpy.ndarray
            Respectively returns the image as bytes or as a tensor
            representation.
        """
        bytess = cls.get(uri)

        if cls.is_image(bytess):
            if as_tensor:
                tensor = cls.encoded_to_tensor(bytess)

                last_dimensions = tensor.get_shape()[-1]

                if last_dimensions == 1:
                    tensor = image_ops.grayscale_to_rgb(tensor)
                elif last_dimensions == 4:
                    tensor = io_ops.decode_png(bytess, channels=3)

                tensor = tensorflow.dtypes.cast(
                    tensor,
                    tensorflow.float32
                )

                return tensor

            return bytess

        raise FileIsNotImage(f'{uri} is not recognized as a valid image file.')

    @classmethod
    def is_animated_gif(cls, uri):
        """Tests if the content is an animated GIF image.

        Parameters
        ----------
        uri: str
            Place where the animated GIF is stored.

        Returns
        -------
        bool:
            Describing if ``tensor`` represents a GIF animation or not.
        """
        tensor = cls.get_image(uri, as_tensor=True)

        if tensor.ndim == 4 and tensor.get_shape()[0] != 1:
            return True

        return False

    @classmethod
    def mean_gif(cls, uri):
        """Transforms an animated GIF to an static representation
        (based on the mean image).

        Parameters
        ----------
        uri: str
            Place where the animated GIF is stored.

        Returns
        -------
        numpy.ndarray:
            A new representation for the GIF (animated) image.
        """
        if cls.is_animated_gif(uri):
            return tensorflow.math.reduce_mean(
                cls.get_image(uri, as_tensor=True),
                axis=0
            )

        print(
            f'File {uri} is not an animated gif. Returning image itself.'
        )

        return cls.get_image(uri, as_tensor=True)

    @classmethod
    def get_image_base64(cls, uri):
        """Loads an image as a base64 encoded string.

        Parameters
        ----------
        uri: str
            Location where the image is stored.

        Returns
        -------
        bytes:
            The base64 encoded image format.
        """
        return b64encode(cls.get_image(uri, as_tensor=False))

    def scan_images(self, uri):
        """
        Scans a source looking for images.

        Parameters
        ----------
        uri: str
            Location where the source is stored.

        Yields
        ------
        str:
            The validated image file on the passed ``uri``.
        """
        for ffile in self.scan(uri):
            file_bytes = self.get(ffile)

            if self.is_image(file_bytes):
                yield ffile

    @classmethod
    def get_image_bytes_to_base64(cls, image_bytes):
        """Returns the image representation in base64, from its
        bytes representation.

        Parameters
        ----------
        image_bytes: bytes
            Bytes representation of some image.

        Raises
        ------
        FileIsNotImage:
            If the file is not recognized as an image.

        Returns
        -------
        bytes:
            The base64 encoded image format.
        """
        if cls.is_image(image_bytes):
            return b64encode(image_bytes)

        raise FileIsNotImage('Input image was not recognized as a valid one.')

    @staticmethod
    def save_image(path, bytess):
        """Saves an image to a defined path.

        Parameters
        ----------
        path: str
            Local to save the image.

        bytess: bytes
            Bytes representing an image.
        """
        io_ops.write_file(path, bytess)

    @classmethod
    def size(cls, uri, new_size=None, keep_aspect=False):
        """Returns the current image dimensions or resize it.

        Parameters
        ----------
        uri: str
            Description of where the image are located.

        new_size: tuple of ints
            The new intended dimension of the image.

        keep_aspect: bool
            If the image proportions should be preserved.
            or not.

        Returns
        -------
        tuple or numpy.ndarray:
            Respectively with current image dimensions or if new_size is
            ``True``, a resized image, if ``new_size`` parameter was passed
            through.
        """
        if new_size:
            return cls.resize_tensor(
                cls.get_image(uri, as_tensor=True),
                new_size,
                keep_aspect=keep_aspect
            )

        return cls.get_image(uri, as_tensor=True).shape[:2]

    @staticmethod
    def resize_tensor(tensor, size, keep_aspect=False):
        """Resize a tensor representing an image.

        Parameters
        ----------
        tensor: numpy.ndarray
            An image transformed to a raw tensor.

        size: tuple of ints
            The new size.

        Returns
        -------
        numpy.ndarray
            A resized tensor based on ``size`` parameter.
        """
        return image_ops.resize(
            tensor, size, method=image_ops.ResizeMethod.BILINEAR,
            antialias=True,
            preserve_aspect_ratio=keep_aspect
        )

    @classmethod
    def compress(cls, tensor, as_tensor=False):
        """Compress the tensor using the JPEG algorithm.

        Parameters
        ----------
        tensor: numpy.ndarray
            A tensor representing an image

        as_tensor: bool
            If the new compressed JPEG image should be
            returned as a ``numpy.ndarray``.

        Returns
        -------
        bytes or numpy.ndarray:
            An encoded image, in bytes or a tensor in numpy.ndarray
        """
        if tensor.dtype is not tensorflow.uint8:
            tensor = tensorflow.dtypes.cast(
                tensor,
                tensorflow.uint8
            )

        compressed = image_ops.encode_jpeg(
            tensor,
            quality=80,
            progressive=True,
            optimize_size=True,
            chroma_downsampling=True
        )

        if as_tensor:
            return cls.encoded_to_tensor(compressed)

        return compressed

    @staticmethod
    def os_supports_compression():
        """Whether the underlying OS supports compression.

        Returns
        -------
        bool:
            True if the OS supports compression, False otherwise.
        """
        compression_supported = True

        if system() == 'Windows':
            compression_supported = False

        return compression_supported



"""
interface Module.

Create a web interface to query images and see results
based on indexed images on database.
"""
import lzma
import webbrowser
import socketserver
from urllib.parse import urlparse, parse_qs
from http.server import SimpleHTTPRequestHandler

import termcolor

from pupyl.duplex.file_io import FileIO
from pupyl.search import PupylImageSearch


TEMPLATE = lzma.decompress(
    b'\xfd7zXZ\x00\x00\x04\xe6\xd6\xb4F\x02\x00!\x01\x0c\x00\x00\x00'
    b'\x8f\x98A\x9c\xe0\x04T\x02=]\x00\x1e\x08E\x06\xd0\xef\xa8(\x17'
    b'Y\xd9\xa9\xf5\xea\xa7%@j6\\\xeb\x9c\xe5Q9\xfb__\x00W\xfe\x9c\xbd'
    b"\xbc\xb7\x81T\x95\xde\xe59\xdbd[\x8aWP'\xa1\xde\xdf\xb3fV\x05\x00T"
    b' \x84>8\xf1_y\xe1\xa50\xeb\x87\x91\xc0\x9c\x91\xf7\xaa%}\xfa\x98Ju'
    b'\x1f\xbb\x1fva\xdcW\xc9\x94&\xe8\xeb\x1et\xe7m\xa5\xd2\xdc\x0c'
    b'\xe5\x85@\xbc`\xb8\x80\x8b\xb7\t\x1e\x1e\xd71\xa7\xf4'
    b"\x01\xf6\xf4\xb6\xb7\xb3\xbe\r\xef\x01\xb7\xe8\x13\xcd'I"
    b'o\x8f\xde\xc2\xa9\xc8\x8a\xbd\x8e:\x00\xd3\x96\xfe\x9c?\xe3\xd2U#'
    b'\x88\x11\x9c\xfd3g\x8c\xfb\xd4\xb9\xd0\xc9"\xfd\xf5Nob\xd8\x12'
    b'\xa7\x91\x16&\x85\x9e\xe9\xc3sZ6\xb2\n\xc2\xdb$\xab\xe9\xf6}'
    b'l\xed\x06\xd1\x8f\x92\xd3\x82\xed\x91#|\xa2{\x7f\xfa'
    b"\xe6\xae\xfb\x05\xaeI\xa8\xa0\x1b\xca?\x06\xca%P\xea\x9c'\xd7\xce"
    b'\x7f\x8b`zB(|\xe0n\xfb\xdb\xdb\x84\xf3\x1e\xa7\xff\x0f7\x83'
    b'\xcd\xcc\xe39\xef\xe4A\xf50\\\xe9>\x18\xa61P\xce\x98>\x16'
    b'\xc9s\xb3\xa3\xb5\xb6\x84Nv\xaf\x07\x1d\x0b\xe0\xb2\x82\x06C`K'
    b'\x08/TC\xbfr\xb9\xda"\xff\xc9\xa9!=?\xd9`\xfaN,\x01\'t\xc0:m\r1'
    b'\xfa)\x88\xf9\xb6\x15\xaa\x04\x85\xa4\xec\xa8\xb5\x06a}'
    b'=\xb5\xca\x8c>\x0e\x18\x04\xec\xbe\xea1\x9c?\xf5\x8c\xe7\xe36k'
    b'm\n\xb2\xadu\n\xb7\xab sD\x98\xb3\x8fJ\xa5\x0c\x06\xeb5'
    b'\xf6b\xcc\xd0\xbd\x8a#&Z+\xbb-\xe0%\xff\x97<\xaa\xbbL\\\x9c;\x95'
    b'\x1a1Ifl\xbeN\x01\x98\xfb\xa1-N\x9f\xb1q\xb4\x92J\x7f\xf8\xcdf\xf9'
    b'\xeb\xdbM\x89\xd9\xae)\xef\xc7\xbbK\x8c\x1c%\x9e\x90'
    b'\xc1\xd7\xd1\xccM\x9bB\xcf\xa9\nD4>\x0f\x9d\x9fe:\x92\x93'
    b'8\xd7m\xba\xf4&a\x9e\x82D(\x13\xc7\xca\xc4w\xaa\xa0\x17\xf0'
    b'\xc2\x9f\xdb\xd9\xd69*~\xec\xb3\x84\xd3\xc6\x92u\x0c\xf3h\x1b\x9b'
    b'\xab\x9e-\x80]\xbam\xbe\xb0\xddJ\xd6\xda\xc5\x94$\xe88\x00\xec'
    b'i\xa3H\x1a\\\xafvZ}l\x11%\n2\n\xb0\x0b\xf8C\xef\xf1\x06\xdd\xac'
    b'\xf4\xae\xbeh\x96\xea<\x92\xc55\xef\x900\x05\xb3\xf8b\x89\x9bV'
    b'A\xc4\x8c\rZ\x1d\xacB\x16\xf2\xdb\x92\x00\x00\x00\x00i\x99t\xd5'
    b"\xd3\xbb 5\x00\x01\xd9\x04\xd5\x08\x00\x00\xe9'}\x1c\xb1\xc4g\xfb"
    b'\x02\x00\x00\x00\x00\x04YZ'
).decode()


def serve(data_dir=None, port=8080, **kwargs):
    """Starts the web server.

    Parameters
    ----------
    data_dir: str
        The path to the data directory. If not set, will try to find inside
        the default temporary file.

    port: int
        Defines the network port which the web server will start listening.

    extreme_mode: bool
        Optional mode for general acceleration.
    """
    if not data_dir:
        data_dir = FileIO.pupyl_temp_data_dir()

    if kwargs.get('extreme_mode') is None:
        _extreme_mode = True
    else:
        _extreme_mode = kwargs['extreme_mode']

    pupyl_image_search = PupylImageSearch(data_dir, extreme_mode=_extreme_mode)

    class RequestHandler(SimpleHTTPRequestHandler):
        """A web request handler."""

        _data_dir = data_dir

        def __init__(self, request, client_address, server):
            SimpleHTTPRequestHandler.__init__(
                self, request, client_address, server
            )

        def do_GET(self):
            """Handler for GET request methods."""

            query_image = None

            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()

            query_string = parse_qs(urlparse(self.path).query)

            query_list = query_string.get('uri', None)

            image_tags = self.images(query_list)

            if query_list:
                query_image_base64 = pupyl_image_search.image_database.\
                    get_image_base64(query_list[0])

                query_image = '<img class="img-thumbnail" ' +\
                    'src="data:image/jpg;base64, ' +\
                    f'{query_image_base64.decode()}" alt="&#129535; Pupyl">' +\
                    '<figcaption class="figure-caption">' +\
                    'Query image used in the search.</figcaption>'

            self.wfile.write(
                bytes(
                    TEMPLATE.format(
                        images=image_tags,
                        query=query_image if query_image else ''),
                    'utf-8'
                )
            )

        @staticmethod
        def _format_metadata(metadata):
            """Formats metadata information

            Parameters
            ----------
            metadata: dict
                The information to be formatted.
            """
            formatted = ''

            for key, value in metadata.items():
                key = key.replace('_', ' ').title()

                formatted += f'<b>{key}</b>: {value} <br> '

            return formatted

        def images(self, query_uri=None, top=None):
            """Returns image tags from database.

            Parameters
            ----------
            query_uri: str
                Location where the query image is stored.

            top: int
                How many results should be returned from some search request.

            Returns
            -------
            str:
                With base64 strings related to images inside database.

            Raises
            ------
            OSError:
                If some problem happened when starting the server.
            KeyboardInterrupt:
                If a command to stop the server is issued (like CTRL+c).
            """

            image_tags = ''
            img_src = '<figure class="figure">' + \
                '<a href="data:image/jpg;base64, {image_b64}">' + \
                '<img class="img-fluid border"' + \
                'src="data:image/jpg;base64, {image_b64}" ' + \
                'alt="&#129535; Pupyl"></a> ' + \
                '<figcaption class="figure-caption">' + \
                '{figure_caption}</figcaption></figure>'

            top = top if top else 24

            if query_uri:
                query_uri = query_uri[0]

                for result in pupyl_image_search.search(query_uri, top=top):
                    result = int(result)

                    image = pupyl_image_search.image_database.\
                        get_image_bytes_to_base64(
                            pupyl_image_search.image_database.
                            load_image(result)
                        ).decode('utf-8')

                    metadata = self._format_metadata(
                        pupyl_image_search.image_database.
                        load_image_metadata(result)
                    )

                    image_tags += img_src.format(
                        image_b64=image,
                        figure_caption=metadata
                    )

                return image_tags

            for index, image in pupyl_image_search.image_database.list_images(
                return_ids=True,
                top=9
            ):
                image_base64 = pupyl_image_search.image_database.\
                    get_image_base64(
                        image
                    ).decode('utf-8')

                metadata = self._format_metadata(
                    pupyl_image_search.image_database.
                    load_image_metadata(index)
                )

                image_tags += img_src.format(
                    image_b64=image_base64,
                    figure_caption=metadata
                )

            return image_tags

    if not port:
        port = 8080

    try:
        with socketserver.TCPServer(('', port), RequestHandler) as httpd:
            print(
                termcolor.colored(
                    f'Server listening on port {port}.',
                    color='green',
                    attrs=['bold']
                )
            )

            webbrowser.open_new_tab(f'http://localhost:{port}')
            httpd.serve_forever()

    except OSError:
        print(
            termcolor.colored(
                f'Port {port} already in use. Trying {port + 1}...',
                color='red',
                attrs=['bold']
            )
        )

        serve(data_dir=data_dir, port=port + 1)
    except KeyboardInterrupt:
        print('🧿 Pupyl says bye.')



""" Manage cli arguments"""

import os
import json
import argparse
from pathlib import Path

import termcolor

from pupyl.search import PupylImageSearch
from pupyl.web import interface


PUPYL_HOME_FOLDER = os.path.join(Path.home(), 'pupyl')


class PupylCommandLineInterface:
    """Implements the command line interface for pupyl."""

    @staticmethod
    def parsers():
        """Creates parsers and subparsers for pupyl's CLI.

        Returns
        -------
        argparse.ArgumentParser:
            Parsed arguments from command line.
        """

        parser = argparse.ArgumentParser(
            prog='pupyl',
            description="""
            🧿 Manage pupyl CLI arguments. pupyl is a really
            fast image search library which you can index your
            own (millions of) images and find similar images in milliseconds.
            """,
            epilog="""
            👥 Contribute to pupyl on https://github.com/policratus/pupyl
            """
        )

        parser.add_argument(
            '--data_dir',
            type=str,
            default=PUPYL_HOME_FOLDER,
            help='data directory for database assets.'
        )

        sub_parsers = parser.add_subparsers(dest='options')

        # Index parser
        index_parser = sub_parsers.add_parser(
            'index', help='indexes images into the database.'
        )
        index_parser.add_argument(
            'input_images',
            help='data directory containing image files to index.'
        )

        sub_parsers.add_parser(
            'serve',
            help='creates a web interface to interact with the database.'
        )

        # Search parser
        search_parser = sub_parsers.add_parser(
            'search', help='search inside a database for similar images.'
        )
        search_parser.add_argument(
            'query', help='URI of an image to use as query.'
        )
        search_parser.add_argument(
            '--top', type=int, default=10, metavar='n',
            help='filters how many results to show.'
        )
        search_parser.add_argument(
            '--metadata', action='store_true',
            help='returns metadata instead of image ids.'
        )

        # Export parser
        export_parser = sub_parsers.add_parser(
            'export',
            help='search inside database, but export result files to a '
            'directory.'
        )
        export_parser.add_argument(
            'query',
            help='URI of an image to use as query.'
        )
        export_parser.add_argument(
            'output', help='directory to export search results as images.'
        )
        export_parser.add_argument(
            '--top', type=int, default=10, metavar='n',
            help='filters how many results to show.'
        )

        export_exclusive_group = export_parser.add_mutually_exclusive_group()
        export_exclusive_group.add_argument(
            '--keep_ids', action='store_true',
            help='should the original image ids must be preserved or not.'
        )
        export_exclusive_group.add_argument(
            '--keep_names', action='store_true',
            help='should the original image names must be preserved or not.'
        )

        return parser

    @classmethod
    def argument_parser(cls, **kwargs):
        """Renders the command line parsers.

        Parameters
        ----------
        commands: list
            The optional list of parameters to parse.

        Returns
        -------
        argparse.Namespace:
            A namespace with parsed command line arguments whithin.
        """

        commands = kwargs.get('commands')

        if commands:
            return cls.parsers().parse_args(commands)

        return cls.parsers().parse_args()


def pupyl():
    """CLI entry point."""

    cli = PupylCommandLineInterface()
    args = cli.argument_parser()

    if args.data_dir == PUPYL_HOME_FOLDER and args.options:
        print(
            termcolor.colored(
                "Since the argument --data_dir wasn't filled,",
                color='cyan'
            ),
            termcolor.colored(
                f'creating pupyl assets on {args.data_dir}', color='cyan'
            )
        )
    elif not args.options:
        cli.parsers().print_help()

    pupyl_search = PupylImageSearch(data_dir=args.data_dir)

    if args.options == 'index':
        pupyl_search.index(
            args.input_images
        )

    elif args.options == 'serve':
        interface.serve(data_dir=args.data_dir)

    elif args.options == 'search':
        for result in pupyl_search.search(
            query=args.query, return_metadata=args.metadata, top=args.top
        ):
            if isinstance(result, dict):
                print(
                    termcolor.colored(
                        json.dumps(result, indent=4), color='green'
                    )
                )
            else:
                print(termcolor.colored(result, color='green'))

    elif args.options == 'export':
        pupyl_search.indexer.export_results(
            args.output, pupyl_search.search(args.query, top=args.top),
            args.keep_ids, args.keep_names
        )
    else:
        cli.parsers().print_help()



"""Factory for image feature extraction."""

import os
import warnings
from enum import Enum, auto

import termcolor

from pupyl.verbosity import quiet_tf
quiet_tf()

import numpy
import tensorflow
import keras.backend as backend
import keras.applications as networks

from pupyl.duplex.image import ImageIO
from pupyl.embeddings import exceptions


class Characteristics(Enum):
    """Describes high level characteristics of complex feature extractors.

    Notes
    -----
    The currently supported characteristics are:

    ``MINIMUMWEIGHT_FAST_SMALL_PRECISION # MobileNet``

    ``LIGHTWEIGHT_FAST_SMALL_PRECISION # ResNet50V2``

    ``LIGHTWEIGHT_FAST_SHORT_PRECISION # ResNet101V2``

    ``LIGHTWEIGHT_QUICK_SHORT_PRECISION # DenseNet169``

    ``MEDIUMWEIGHT_QUICK_GOOD_PRECISION # DenseNet201``

    ``MIDDLEWEIGHT_QUICK_GOOD_PRECISION # InceptionV3``

    ``MIDDLEWEIGHT_SLOW_GOOD_PRECISION # Xception``

    ``HEAVYWEIGHT_SLOW_GOOD_PRECISION # EfficientNetV2M``

    ``HEAVYWEIGHT_SLOW_HUGE_PRECISION # EfficientNetV2L``
    """
    # MobileNet
    MINIMUMWEIGHT_FAST_SMALL_PRECISION = auto()
    # ResNet50V2
    LIGHTWEIGHT_FAST_SMALL_PRECISION = auto()
    # ResNet101V2
    LIGHTWEIGHT_FAST_SHORT_PRECISION = auto()
    # DenseNet169
    LIGHTWEIGHT_QUICK_SHORT_PRECISION = auto()
    # DenseNet201
    MEDIUMWEIGHT_QUICK_GOOD_PRECISION = auto()
    # InceptionV3
    MIDDLEWEIGHT_QUICK_GOOD_PRECISION = auto()
    # Xception
    MIDDLEWEIGHT_SLOW_GOOD_PRECISION = auto()
    # EfficientNetV2M
    HEAVYWEIGHT_SLOW_GOOD_PRECISION = auto()
    # EfficientNetV2L
    HEAVYWEIGHT_SLOW_HUGE_PRECISION = auto()

    @staticmethod
    def by_name(name):
        """Returns a characteristic by its name.

        Parameters
        ----------
        name: str
            String representation of enumerator

        Raises
        ------
        UnknownCharacteristicsName:
            If ``name`` is unknown.

        Returns
        -------
        enum:
            The corresponding characteristic.
        """
        try:
            return Characteristics[name]
        except KeyError as key_error:
            raise exceptions.UnknownCharacteristicsName(
                f'Characteristic with name {name} is unknown.'
            ) from key_error

    @staticmethod
    def by_value(value):
        """Returns a characteristic by its value.

        Parameters
        ----------
        value: int
            Integer representing a characteristic

        Raises
        ------
        UnknownCharacteristicsValue:
            If ``value`` is unknown.

        Returns
        -------
        enum:
            The corresponding characteristic.
        """
        try:
            return Characteristics(value)
        except ValueError as value_error:
            raise exceptions.UnknownCharacteristicsValue(
                f'Characteristic with value {value} is unknown.'
            ) from value_error


class Extractors(ImageIO):
    """Pretrained CNNs for embedding generation."""

    def __init__(self, characteristics, extreme_mode=True):
        """
        Creates embedding extractors.

        Parameters
        ----------
        characteristics: Enum
            Describing the intended characteristics to transform images
            into its underlying embeddings.

        extreme_mode: bool
            Should the extreme mode (faster execution but not gentle with
            memory) be enabled or disabled?
        """
        self._characteristics = characteristics
        self._extreme_mode = extreme_mode

        self.acceleration_discovery()

        self.converter, self.network = self._infer_network()
        self.image_input_shape = self.network.input_shape[1:3]
        self._features_output_shape = self.network.output_shape[1]

    def __enter__(self):
        """Opens Extractors context."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close Extractors context."""
        del exc_type, exc_val, exc_tb

        backend.clear_session()

    @property
    def output_shape(self):
        """Getter for property output_shape.

        Describes the output shape of extracted features.

        Returns
        -------
        int:
            The output shape for the picked characteristics.
        """
        return self._features_output_shape

    def acceleration_discovery(self):
        """Performs a hardware processing acceleration discovery.

        Most NVIDIA® GPUs supported (through ``CUDA``), which results on
        faster embeddings extraction.
        """
        # If GPU(s) device(s) are found, avoid excessive VRAM consumption
        for gpu in tensorflow.config.list_physical_devices('GPU'):
            print(
                termcolor.colored(
                    'GPU found! Using it for embeddings '
                    'extraction acceleration.',
                    color='blue',
                    attrs=['bold']
                )
            )

            if not self._extreme_mode:
                try:
                    tensorflow.config.experimental.set_memory_growth(gpu, True)
                except RuntimeError:
                    warnings.warn(
                        termcolor.colored(
                            'Trying to define memory growth for an '
                            'already initialized GPU',
                            color='blue',
                            attrs=['bold']
                        ),
                        ResourceWarning
                    )

    def _infer_network(self):
        """Translates a characteristic to a network architecture.

        Raises
        ------
        UnknownCharacteristics:
            If the characteristics passed through are unknown.

        Returns
        -------
        tuple:
            With preprocessors and complete CNN architecture.
        """
        # Networks general configurations
        weights = 'imagenet'
        pooling = 'max'
        include_top = False
        small_input_shape = (224, 224, 3)
        middle_input_shape = (299, 299, 3)
        large_input_shape = (480, 480, 3)

        if self._characteristics is \
                Characteristics.MINIMUMWEIGHT_FAST_SMALL_PRECISION:
            return networks.mobilenet.preprocess_input, \
                networks.mobilenet.MobileNet(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=small_input_shape
                )

        if self._characteristics is \
                Characteristics.LIGHTWEIGHT_FAST_SMALL_PRECISION:
            return networks.resnet_v2.preprocess_input, \
                networks.resnet_v2.ResNet50V2(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=small_input_shape
                )

        if self._characteristics is \
                Characteristics.LIGHTWEIGHT_FAST_SHORT_PRECISION:
            return networks.resnet_v2.preprocess_input, \
                networks.resnet_v2.ResNet101V2(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=small_input_shape
                )

        if self._characteristics is \
                Characteristics.LIGHTWEIGHT_QUICK_SHORT_PRECISION:
            return networks.densenet.preprocess_input, \
                networks.densenet.DenseNet169(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=small_input_shape
                )

        if self._characteristics is \
                Characteristics.MEDIUMWEIGHT_QUICK_GOOD_PRECISION:
            return networks.densenet.preprocess_input, \
                networks.densenet.DenseNet201(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=small_input_shape
                )

        if self._characteristics is \
                Characteristics.MIDDLEWEIGHT_QUICK_GOOD_PRECISION:
            return networks.inception_v3.preprocess_input, \
                networks.inception_v3.InceptionV3(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=middle_input_shape
                )

        if self._characteristics is \
                Characteristics.MIDDLEWEIGHT_SLOW_GOOD_PRECISION:
            return networks.xception.preprocess_input, \
                networks.xception.Xception(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=middle_input_shape
                )

        if self._characteristics is \
                Characteristics.HEAVYWEIGHT_SLOW_GOOD_PRECISION:
            return networks.efficientnet_v2.preprocess_input, \
                networks.efficientnet_v2.EfficientNetV2M(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=large_input_shape,
                    include_preprocessing=True
                )

        if self._characteristics is \
                Characteristics.HEAVYWEIGHT_SLOW_HUGE_PRECISION:
            return networks.efficientnet_v2.preprocess_input, \
                networks.efficientnet_v2.EfficientNetV2L(
                    weights=weights,
                    pooling=pooling,
                    include_top=include_top,
                    input_shape=large_input_shape,
                    include_preprocessing=True
                )

        raise exceptions.UnknownCharacteristics(
            f'Characteristic {self._characteristics}'
            ' is invalid.'
        )

    def preprocessor(self, uri):
        """Image preprocessing methods, suitable for posterior network
        ingestion. It may include image resizing, normalization, among others.

        Parameters
        ----------
        uri: str
            Location of the image to be preprocessed.

        Returns
        -------
        numpy.ndarray
            Containing the processed image
        """
        if self.is_animated_gif(uri):
            tensor = self.resize_tensor(
                self.mean_gif(uri), self.image_input_shape
            )
        else:
            tensor = self.size(uri, self.image_input_shape)

        return tensorflow.expand_dims(self.converter(tensor), axis=0)

    def extract(self, uri):
        """Converts image uri to its embeddings.

        Parameters
        ----------
        uri: str
            Location of the image to be converted to a embedding.

        Returns
        -------
        numpy.ndarray
            1D tensor with extracted features.
        """
        return tensorflow.squeeze(
            self.network.predict(self.preprocessor(uri), verbose=0)
        )

    def extract_save(self, path, uri):
        """Extracts features from an image, saving to disk after all.

        Parameters
        ----------
        path: str
            Where to store the tensor.

        uri: str
            Location of the image to be converted to a embedding.
        """
        if not os.path.exists(path):
            self.save(path, self.extract(uri))

    @staticmethod
    def save(path, tensor):
        """Writes down a ``tensor`` referencing ``index``.

        Parameters
        ----------
        path: str
            Where to store the tensor.

        tensor: Tensor
            The tensor itself, to be saved.
        """
        numpy.save(path, tensor.numpy(), allow_pickle=False, fix_imports=False)

    @staticmethod
    def load(path):
        """Loads up a ``tensor`` referencing ``index``.

        Parameters
        ----------
        path: str
            Where to load from the tensor.

        Returns
        -------
        numpy.ndarray
            Tensor loaded back again.
        """
        return numpy.load(path, mmap_mode=None, allow_pickle=False)



"""Exceptions regarding embeddings module operations."""


class Error(Exception):
    """Base exception class."""


class FileTypeNotSupportedYet(Error):
    """Raised when an unknown file type was discovered."""


class FileIsNotImage(Error):
    """Raised when the analyzed file is not an image."""


class FileScanNotPossible(Error):
    """Raised when the analyzed file is not suitable for scan."""



"""Unit tests for the command line interface."""
from argparse import Namespace
from unittest import TestCase

from pupyl.cli import PupylCommandLineInterface


CLI = PupylCommandLineInterface()


class TestCases(TestCase):
    """Unit test for special cases."""

    def test_argument_parser_without_parameters(self):
        """Unit test for method argument_parser, without params. case."""

        with self.assertRaises(SystemExit):

            _ = CLI.argument_parser()


def test_parser_index_case():
    """Unit test for the index subcommand with all arguments."""
    sub_command = 'index'
    input_path = 'path/to/images'
    data_dir = 'path/to/assets'

    expected_vars = {
        'data_dir': data_dir,
        'input_images': input_path,
        'options': sub_command,
    }

    args = CLI.argument_parser(
        commands=['--data_dir', data_dir, sub_command, input_path]
    )

    assert vars(args) == expected_vars


def test_parser_serve_case():
    """Unit test for the serve subcommand with all arguments."""
    sub_command = 'serve'
    data_dir = 'path/to/assets'

    expected_vars = {
        'data_dir': data_dir,
        'options': sub_command,
    }

    args = CLI.argument_parser(
        commands=['--data_dir', data_dir, sub_command]
    )

    assert vars(args) == expected_vars


def test_parser_export_case():
    """Unit test for the export subcommand with all arguments."""
    sub_command = 'export'
    query_image = 'path/to/query/image'
    output_dir = 'path/to/output/dir'
    data_dir = 'path/to/assets'
    top = 10

    expected_vars = {
        'data_dir': data_dir,
        'query': query_image,
        'output': output_dir,
        'options': sub_command,
        'top': top,
        'keep_ids': False,
        'keep_names': False
    }

    args = CLI.argument_parser(
        commands=['--data_dir', data_dir, sub_command, query_image, output_dir]
    )

    assert vars(args) == expected_vars

    expected_vars['keep_ids'] = True

    args = CLI.argument_parser(
        commands=[
            '--data_dir', data_dir, sub_command, query_image, output_dir,
            '--keep_ids'
        ]
    )

    assert vars(args) == expected_vars


def test_argument_parser_parameters():
    """Unit test for method argument_parser, passing params. case."""
    sub_command = 'index'
    input_path = 'path/to/images'
    data_dir = 'path/to/assets'

    test_namespace = Namespace(
        data_dir=data_dir,
        input_images=input_path,
        options=sub_command
    )

    test_arguments_parsed = CLI.argument_parser(
        commands=['--data_dir', data_dir, sub_command, input_path]
    )

    assert test_namespace == test_arguments_parsed



"""Controls verbosity of some modules."""

import os
import logging


def quiet_tf():
    """Suppress some messages from ``tensorflow``."""
    logging.getLogger('tensorflow').setLevel(
        logging.ERROR
    )
    os.environ['KMP_AFFINITY'] = 'noverbose'
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    import tensorflow

    tensorflow.get_logger().setLevel(logging.ERROR)
    tensorflow.autograph.set_verbosity(0, alsologtostdout=False)



"""Unit tests related to indexer.facets module."""
import os
import tempfile
import warnings
from unittest import TestCase

import numpy

from pupyl.indexer.facets import Index
from pupyl.indexer.exceptions import FileIsNotAnIndex, IndexNotBuildYet, \
    NoDataDirForPermanentIndex, DataDirDefinedForVolatileIndex, \
    NullTensorError, TopNegativeOrZero, EmptyIndexError, ExportIdsAndNames
from pupyl.duplex.file_io import FileIO
from pupyl.search import PupylImageSearch
from pupyl.embeddings.features import Extractors, Characteristics


TEST_INDEX_PATH = os.path.abspath('tests')
TEST_INDEX_SEARCH_PATH = os.path.join(TEST_INDEX_PATH, 'test_search')
TEST_INDEX_INVALID_FILE = os.path.join(TEST_INDEX_PATH, 'test_index')
TEST_INDEX_EXPORT = os.path.join(TEST_INDEX_PATH, 'test_index_export')
TEST_EMPTY_INDEX = os.path.join(TEST_INDEX_PATH, 'test_empty_index')
TEST_CHECK_UNIQUE = os.path.join(TEST_INDEX_PATH, 'test_check_unique')
TEST_ANIMATED_GIF = os.path.join(TEST_INDEX_PATH, 'test_gif.gif')
TEST_VECTOR_SIZE = 128

INDEX = Index(TEST_VECTOR_SIZE, TEST_INDEX_PATH)
ALL_ITEMS = list(INDEX.items())
ALL_VALUES = list(INDEX.values())


class TestCases(TestCase):
    """Unit tests over special cases."""

    def test___next_stop_iteration(self):
        """Unit test for method __next__, StopIteration case."""
        with self.assertRaises(StopIteration):
            index = Index(TEST_VECTOR_SIZE, TEST_INDEX_PATH)
            while True:
                next(index)

    def test_open_invalid_file(self):
        """Unit test for method __init__, invalid file case."""
        with self.assertRaises(FileIsNotAnIndex):
            Index(TEST_VECTOR_SIZE, TEST_INDEX_INVALID_FILE)

    def test_remove_new_file(self):
        """Unit test for method remove, new file case."""
        with self.assertRaises(IndexNotBuildYet):
            with Index(size=8, volatile=True) as index:
                index.remove(0)

    def test_data_dir_is_not_a_dir(self):
        """Unit test for __init__, data dir. as file case."""
        with self.assertRaises(OSError):
            Index(
                TEST_VECTOR_SIZE,
                os.path.join(TEST_INDEX_PATH, 'pupyl.index')
            )

    def test_no_data_dir_for_perm_file(self):
        """Unit test for __init__, no data dir for perm. file case."""
        with self.assertRaises(NoDataDirForPermanentIndex):
            Index(TEST_VECTOR_SIZE, data_dir=None)

    def test_data_dir_for_volatile_file(self):
        """Unit test for __init__, data dir for volatile file case."""
        with self.assertRaises(DataDirDefinedForVolatileIndex):
            Index(
                TEST_VECTOR_SIZE,
                data_dir=os.path.join(TEST_INDEX_PATH, 'pupyl.index'),
                volatile=True
            )

    def test_raises_null_tensor_error(self):
        """Unit test for append, null tensor case."""
        with self.assertRaises(NullTensorError):
            with Index(TEST_VECTOR_SIZE, volatile=True) as index:
                index.append(numpy.zeros(shape=TEST_VECTOR_SIZE))

    def test_remove_index_error(self):
        """Unit test for remove, index error case."""
        with self.assertRaises(IndexError):
            INDEX.remove(999)

    def test_negative_top(self):
        """Unit test for top parameter, negative case."""

        with self.assertRaises(TopNegativeOrZero):
            with Index(4, TEST_INDEX_SEARCH_PATH) as index:
                _ = [*index.group_by(top=-1)]

    def test_empty_index(self):
        """Unit test for an empty index."""
        with self.assertRaises(EmptyIndexError):
            with Index(1, TEST_EMPTY_INDEX) as index:
                _ = [*index.group_by()]

    def test_export_exclusive_options(self):
        """Unit test for exclusive options."""
        test_vector_size = 1024

        with self.assertRaises(ExportIdsAndNames):
            with tempfile.TemporaryDirectory() as temp_dir:
                test_search = PupylImageSearch(temp_dir)

                test_search.index(TEST_INDEX_EXPORT)

                with tempfile.TemporaryDirectory() as new_temp_dir:
                    with Index(test_vector_size, data_dir=temp_dir) as index:
                        index.export_results(
                            new_temp_dir,
                            test_search.search(
                                os.path.join(TEST_INDEX_EXPORT, '1.jpg')
                            ),
                            keep_ids=True,
                            keep_names=True
                        )


def test_append_check_unique():
    """Unit test for method append, check unique case."""
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter('always')

        with Extractors(
            Characteristics.MINIMUMWEIGHT_FAST_SMALL_PRECISION
        ) as extractor, \
                Index(extractor.output_shape, volatile=True) as index:
            for image in extractor.scan(TEST_CHECK_UNIQUE):
                index.append(
                    extractor.extract(image),
                    check_unique=True
                )

        assert len(caught_warnings) == 1
        assert issubclass(caught_warnings[-1].category, UserWarning)
        assert str(
            caught_warnings[-1].message
        ) == 'Tensor being indexed already exists in ' + \
            'the database and the check for duplicates ' + \
            'are on. Refusing to store this tensor again.'


def test_open_index():
    """Unit test for index opening."""
    assert INDEX.size == TEST_VECTOR_SIZE


def test_context():
    """Unit test to test context opening and close."""
    with Index(TEST_VECTOR_SIZE, TEST_INDEX_PATH) as index:
        assert isinstance(index, Index)


def test_items():
    """Unit test for method items."""
    for test_item, method_item in zip(ALL_ITEMS, INDEX.items()):
        assert test_item == method_item


def test_values():
    """Unit test for method items."""
    for test_value, method_value in zip(ALL_VALUES, INDEX.values()):
        assert test_value == method_value


def test_items_values():
    """Unit test for method items_values."""
    for items_values, test_item, test_value in zip(
            INDEX.items_values(), ALL_ITEMS, ALL_VALUES
    ):
        method_item, method_value = items_values

        assert test_item == method_item and \
            test_value == method_value


def test_preprocessor_animated_gif():
    """Unit test for method preprocessor, animated GIF case."""
    with Extractors(
        Characteristics.HEAVYWEIGHT_SLOW_HUGE_PRECISION
    ) as extractors:
        assert extractors.preprocessor(TEST_ANIMATED_GIF).shape == \
            (1, 480, 480, 3)


def test___get_item__():
    """Unit test for magic method __getitem__."""
    assert INDEX[0] == ALL_VALUES[0]


def test_reversed__get_item__():
    """Unit test for magic method __getitem__, reversed case."""
    assert INDEX[-1] == ALL_VALUES[-1]


def test___len__():
    """Unit test for magic method __len__"""
    assert len(INDEX) == len(ALL_ITEMS)


def test___iter__():
    """Unit test for magic method __iter__"""
    for test_value, method_value in zip(ALL_VALUES, INDEX):
        assert test_value == method_value


def test___next__():
    """Unit test for magic method __next__"""
    for value in ALL_VALUES:
        assert next(INDEX) == value


def test_size():
    """Unit test for method size."""
    assert INDEX.size == TEST_VECTOR_SIZE


def test_path_property():
    """Unit test for path property."""
    assert INDEX.path == os.path.join(TEST_INDEX_PATH, INDEX.index_name)


def test_trees_property():
    """Unit test for path property."""
    assert INDEX.trees == .01


def test_volatile_property():
    """Unit test for volatile property."""
    assert not INDEX.volatile


def test_trees_property_setter():
    """Unit test for path property."""
    INDEX.trees = 1

    assert INDEX.trees == 1


def test_open_new_index():
    """Unit test for new index opening."""
    with Index(TEST_VECTOR_SIZE, volatile=True) as index:
        assert index.size == TEST_VECTOR_SIZE


def test_append_new_file():
    """Unit test for method append, new file case."""
    with Index(TEST_VECTOR_SIZE, volatile=True) as index:

        test_size_before = len(index)
        new_tensor = numpy.random.normal(size=TEST_VECTOR_SIZE)

        index.append(new_tensor)

        test_size_after = len(index)

        assert test_size_after == test_size_before + 1

        numpy.testing.assert_array_almost_equal(
            index[-1], new_tensor, decimal=7
        )


def test_append_new_created_file():
    """Unit test for method append, created file case."""
    test_size_before = len(INDEX)
    new_tensor = numpy.random.normal(size=TEST_VECTOR_SIZE)

    INDEX.append(new_tensor)

    test_size_after = len(INDEX)

    assert test_size_after == test_size_before + 1

    numpy.testing.assert_array_almost_equal(
        INDEX[-1], new_tensor, decimal=3
    )


def test_remove():
    """Unit test for method remove."""
    index_to_remove = 8

    temp_file = FileIO.safe_temp_file(file_name='pupyl.index')
    temp_dir = os.path.dirname(temp_file)

    with Index(TEST_VECTOR_SIZE, data_dir=temp_dir) as index:
        for _ in range(16):
            index.append(numpy.random.normal(size=TEST_VECTOR_SIZE))

        test_size_before = len(index)

        test_value = index[index_to_remove]

    with Index(TEST_VECTOR_SIZE, data_dir=temp_dir) as index:
        index.remove(index_to_remove)

        assert len(index) == test_size_before - 1

        numpy.testing.assert_raises(
            AssertionError,
            numpy.testing.assert_array_equal,
            test_value,
            index[index_to_remove]
        )


def test_pop():
    """Unit test for method pop."""
    temp_file = FileIO.safe_temp_file(file_name='pupyl.index')
    temp_dir = os.path.dirname(temp_file)

    with Index(TEST_VECTOR_SIZE, data_dir=temp_dir) as index:
        for _ in range(16):
            index.append(numpy.random.normal(size=TEST_VECTOR_SIZE))

        test_size_before = len(index)

        test_value_before = index[-1]

    with Index(TEST_VECTOR_SIZE, data_dir=temp_dir) as index:
        test_value_after = index.pop()

        assert len(index) == test_size_before - 1

        numpy.testing.assert_array_equal(
            test_value_before,
            test_value_after
        )


def test_pop_index():
    """Unit test for method pop, index case."""
    index_to_pop = 4

    temp_file = FileIO.safe_temp_file(file_name='pupyl.index')
    temp_dir = os.path.dirname(temp_file)

    with Index(TEST_VECTOR_SIZE, data_dir=temp_dir) as index:
        for _ in range(16):
            index.append(numpy.random.normal(size=TEST_VECTOR_SIZE))

        test_size_before = len(index)

        test_value_before = index[index_to_pop]

    with Index(TEST_VECTOR_SIZE, data_dir=temp_dir) as index:
        test_value_after = index.pop(index_to_pop)

        assert len(index) == test_size_before - 1

        numpy.testing.assert_array_equal(
            test_value_before,
            test_value_after
        )


def test_index():
    """Unit test for method index."""
    test_position = 0

    with Index(TEST_VECTOR_SIZE, TEST_INDEX_PATH) as index:
        test_value = index[test_position]

        assert index.index(test_value) == 0


def test_search():
    """Unit test for method search."""
    query_array = [-0.48870765, -0.57780915, -0.94986234, -1.90035123]
    expected_search_result = [25, 78]

    with Index(len(query_array), TEST_INDEX_SEARCH_PATH) as index:
        test_result = [*index.search(query_array, results=2)]

    assert expected_search_result == test_result


def test_search_return_distances():
    """Unit test for method search, return_distances=True case."""
    query_array = [-0.48870765, -0.57780915, -0.94986234, -1.90035123]
    expected_search_result = (78, .603)

    with Index(len(query_array), TEST_INDEX_SEARCH_PATH) as index:
        test_result = [
            *index.search(query_array, results=1, return_distances=True)
        ]

    test_result_index = test_result[0][0]
    test_result_value = round(test_result[0][1], 3)

    assert expected_search_result[0] == test_result_index
    assert expected_search_result[1] == test_result_value


def test_group_by():
    """Unit test for method group_by."""
    expected_result = {0: [25]}

    with Index(4, TEST_INDEX_SEARCH_PATH) as index:
        test_result = [*index.group_by(top=1)][0]

    assert expected_result == test_result


def test_group_by_position():
    """Unit test for method group_by."""
    expected_result = [25, 89, 88, 44, 38, 64, 10, 101, 78, 67]

    with Index(4, TEST_INDEX_SEARCH_PATH) as index:
        test_result = [*index.group_by(position=0)][0]

    assert expected_result == test_result


def test_export_group_by():
    """Unit test for method export_group_by method."""
    test_vector_size = 1024

    with tempfile.TemporaryDirectory() as temp_dir:
        test_search = PupylImageSearch(temp_dir)
        test_search.index(TEST_INDEX_EXPORT)

        with tempfile.TemporaryDirectory() as new_temp_dir:
            with Index(test_vector_size, data_dir=temp_dir) as index:
                index.export_by_group_by(new_temp_dir)

                assert os.path.exists(os.path.join(new_temp_dir, '0', '1.jpg'))
                assert os.path.exists(
                    os.path.join(new_temp_dir, '0', 'group.jpg')
                )
                assert os.path.exists(os.path.join(new_temp_dir, '1', '1.jpg'))
                assert os.path.exists(
                    os.path.join(new_temp_dir, '1', 'group.jpg')
                )


def test_export_group_by_position():
    """Unit test for method export_group_by method, position case."""
    test_vector_size = 1024

    with tempfile.TemporaryDirectory() as temp_dir:
        test_search = PupylImageSearch(temp_dir)
        test_search.index(TEST_INDEX_EXPORT)

        with tempfile.TemporaryDirectory() as new_temp_dir:
            with Index(test_vector_size, data_dir=temp_dir) as index:
                index.export_by_group_by(new_temp_dir, position=1)

                assert os.path.exists(os.path.join(new_temp_dir, '1', '1.jpg'))
                assert os.path.exists(
                    os.path.join(new_temp_dir, '1', 'group.jpg')
                )


def test_export_results():
    """Unit test for method export_results."""
    test_vector_size = 1024

    with tempfile.TemporaryDirectory() as temp_dir:
        test_search = PupylImageSearch(temp_dir)
        test_search.index(TEST_INDEX_EXPORT)

        with tempfile.TemporaryDirectory() as new_temp_dir:
            with Index(test_vector_size, data_dir=temp_dir) as index:
                index.export_results(
                    new_temp_dir, test_search.search(
                        os.path.join(TEST_INDEX_EXPORT, '1.jpg')
                    ),
                    keep_ids=True
                )

                assert os.path.exists(os.path.join(new_temp_dir, '1.jpg'))

                index.export_results(
                    new_temp_dir, test_search.search(
                        os.path.join(TEST_INDEX_EXPORT, '1.jpg')
                    ),
                    keep_names=True
                )

                assert os.path.exists(os.path.join(new_temp_dir, '1.jpg'))



"""
Unit tests related to embeddings.features modules
"""
from os.path import abspath
from enum import Enum, auto
from unittest import TestCase

from pupyl.embeddings.exceptions import UnknownCharacteristics, \
    UnknownCharacteristicsValue, UnknownCharacteristicsName
from pupyl.embeddings.features import Characteristics, Extractors


TEST_LOCAL = abspath('tests/test_image.jpg')
UNKNOWN_CHARACTERISTICS = 'UNKNOWN'


class TestCharacteristics(Enum):
    """Defines enumerators to test if all characteristics are described."""
    MINIMUMWEIGHT_FAST_SMALL_PRECISION = auto()
    LIGHTWEIGHT_FAST_SMALL_PRECISION = auto()
    LIGHTWEIGHT_FAST_SHORT_PRECISION = auto()
    LIGHTWEIGHT_QUICK_SHORT_PRECISION = auto()
    MEDIUMWEIGHT_QUICK_GOOD_PRECISION = auto()
    MIDDLEWEIGHT_QUICK_GOOD_PRECISION = auto()
    MIDDLEWEIGHT_SLOW_GOOD_PRECISION = auto()
    HEAVYWEIGHT_SLOW_GOOD_PRECISION = auto()
    HEAVYWEIGHT_SLOW_HUGE_PRECISION = auto()


class TestCases(TestCase):
    """Unit tests over special cases."""

    def test_unknown_network_instance(self):
        """Unit test for an unknown network characteristic."""
        with self.assertRaises(UnknownCharacteristics):
            Extractors(UNKNOWN_CHARACTERISTICS)

    def test_unknown_characteristic(self):
        """Unit test for an unknown characteristic key."""
        with self.assertRaises(KeyError):
            _ = Characteristics[UNKNOWN_CHARACTERISTICS]

    def test_unknown_characteristic_name(self):
        """Unit test for an unknown characteristic name."""
        with self.assertRaises(UnknownCharacteristicsName):
            _ = Characteristics.by_name(UNKNOWN_CHARACTERISTICS)

    def test_unknown_characteristic_value(self):
        """Unit test for an unknown characteristic value."""
        with self.assertRaises(UnknownCharacteristicsValue):
            _ = Characteristics.by_value(999)


def test_characteristics():
    """Unit test for characteristics definition."""
    for characteristic in TestCharacteristics._member_names_:
        assert characteristic in Characteristics._member_names_


def test_by_value():
    """Unit test for characteristic retrieval by value."""
    for test_characteristic in TestCharacteristics:
        assert test_characteristic.name == Characteristics.by_value(
            test_characteristic.value
        ).name


def test_network_instance():
    """Unit test for network instantiation."""
    for characteristic in Characteristics:
        with Extractors(characteristic) as extractor:
            assert extractor._infer_network()[1].name == extractor.network.name


def test_extractors_context_manager():
    """Unit test for context manager Extractors."""
    for characteristic in Characteristics:
        with Extractors(characteristic) as extractors:
            assert isinstance(extractors, Extractors)


def test_preprocessor():
    """
    Unit test for preprocessor method
    """
    for characteristic in Characteristics:
        with Extractors(characteristic) as extractor:
            assert extractor.preprocessor(abspath(TEST_LOCAL)).shape == \
                (1, *extractor.image_input_shape, 3)


def test_extract():
    """
    Unit test for extract method
    """
    for characteristic in Characteristics:
        with Extractors(characteristic) as extractor:
            assert extractor.extract(abspath(TEST_LOCAL)).shape[0] == \
                extractor.network.output_shape[1]


def test_output_shape_property():
    """Unit test for property output_shape."""
    for characteristic in Characteristics:
        with Extractors(characteristic) as extractor:
            assert extractor.output_shape == extractor._features_output_shape



"""Unit tests for main module pupyl."""
import os
import json
from unittest import TestCase
from urllib.error import URLError
from tempfile import gettempdir, TemporaryDirectory

from pupyl.indexer.facets import Index
from pupyl.search import PupylImageSearch
from pupyl.duplex.exceptions import FileIsNotImage
from pupyl.embeddings.features import Extractors, Characteristics


TEST_DATA_DIR = os.path.join(gettempdir(), 'pupyl_tests/')
PUPYL = PupylImageSearch(data_dir=TEST_DATA_DIR, extreme_mode=True)
TEST_SCAN_DIR = os.path.abspath('tests/test_scan/test_csv.csv.gz')
TEST_INVALID_URL = os.path.abspath('tests/test_scan/invalid.csv')
TEST_QUERY_IMAGE = 'https://static.flickr.com/210/500863916_bdd4b8cc5a.jpg'
TEST_INDEX_INVALID = os.path.abspath('tests/test_index_invalid')
TEST_CONFIG_DIR = os.path.abspath('tests/test_index/')


class TestCases(TestCase):
    """Unit tests over special cases."""

    def test_index_invalid_url(self):
        """Unit test for method index, invalid url case."""
        with self.assertRaises(URLError):
            PUPYL.index(TEST_INVALID_URL)

    def test_index_gaps(self):
        """Unit test for method index, gaps in images case."""
        with self.assertRaises(FileIsNotImage):
            with TemporaryDirectory() as temp_dir:
                pupyl_index_gaps = PupylImageSearch(data_dir=temp_dir)
                pupyl_index_gaps.index(TEST_INDEX_INVALID)


def test_index_no_config_file():
    """
    Unit test for PupylImageSearch class initialization,
    without configuration file case.
    """
    with TemporaryDirectory() as temp_dir:
        test_config = PupylImageSearch(data_dir=temp_dir).\
            _index_configuration('r')

        assert not test_config


def test_index_creation_chosen_parameters():
    """
    Unit test for PupylImageSearch class initialization,
    with chosen parameters case.
    """
    with TemporaryDirectory() as temp_dir:
        pupyl_test = PupylImageSearch(
            data_dir=temp_dir,
            import_images=True,
            characteristic=Characteristics.MINIMUMWEIGHT_FAST_SMALL_PRECISION
        )

        assert pupyl_test._import_images and \
            pupyl_test._characteristic == Characteristics.\
            MINIMUMWEIGHT_FAST_SMALL_PRECISION


def test_index_config_file():
    """Unit test for index creation, with config. file case."""
    test_config_file = 'index.json'

    with open(
        os.path.join(TEST_CONFIG_DIR, test_config_file)
    ) as config:
        configurations = json.load(config)

    test_pupyl = PupylImageSearch(TEST_CONFIG_DIR, extreme_mode=False)

    assert configurations['import_images'] == \
        test_pupyl._import_images

    assert configurations['characteristic'] == \
        test_pupyl._characteristic.name


def test_index():
    """Unit test for method index."""
    PUPYL.index(TEST_SCAN_DIR)

    assert os.path.isdir(TEST_DATA_DIR) and \
        os.path.isfile(os.path.join(TEST_DATA_DIR, 'pupyl.index')) and \
        os.path.isfile(os.path.join(TEST_DATA_DIR, '0', '0.jpg'))


def test_characteristic_by_name():
    """Unit test for instantiating a characteristic by its name."""
    test_characteristic = 'MEDIUMWEIGHT_QUICK_GOOD_PRECISION'

    with TemporaryDirectory() as temp_dir:
        test_pupyl = PupylImageSearch(
            data_dir=temp_dir,
            characteristic=test_characteristic
        )

        assert test_pupyl._characteristic.name == test_characteristic


def test_characteristic_by_value():
    """Unit test for instantiating a characteristic by its value."""
    # LIGHTWEIGHT_QUICK_SHORT_PRECISION
    test_characteristic = 4

    with TemporaryDirectory() as temp_dir:
        test_pupyl = PupylImageSearch(
            data_dir=temp_dir,
            characteristic=test_characteristic
        )

        assert test_pupyl._characteristic.value == test_characteristic


def test_index_no_extreme_mode():
    """Unit test for method index, non extreme mode case."""
    pupyl_non_extreme = PupylImageSearch(
        data_dir=TEST_DATA_DIR,
        extreme_mode=False
    )
    pupyl_non_extreme.index(TEST_SCAN_DIR)

    assert os.path.isdir(TEST_DATA_DIR) and \
        os.path.isfile(os.path.join(TEST_DATA_DIR, 'pupyl.index')) and \
        os.path.isfile(os.path.join(TEST_DATA_DIR, '0', '0.jpg'))


def test_pupyl_temp_data_dir():
    """Unit test for instance saving on temporary dir."""
    pupyl_test = PupylImageSearch()

    assert isinstance(pupyl_test, PupylImageSearch)


def test_search():
    """Unit test for method search."""
    expected_length_results = 1

    test_results = PUPYL.search(
        TEST_QUERY_IMAGE,
        top=expected_length_results
    )

    assert len([*test_results]) == expected_length_results


def test_search_distances():
    """Unit test for method search, returning distances case."""
    expected_result_index = 0
    expected_result_distance = .1

    test_results = PUPYL.search(
        TEST_QUERY_IMAGE,
        top=1,
        return_distances=True
    )

    test_results = [*test_results][0]
    test_results_index = test_results[0]
    test_results_distance = round(test_results[1], 1)

    assert test_results_index == expected_result_index
    assert test_results_distance == expected_result_distance


def test_search_non_extreme_mode():
    """Unit test for method search, non-extreme mode case."""
    expected_length_results = 1

    pupyl_non_extreme = PupylImageSearch(
        data_dir=TEST_DATA_DIR,
        extreme_mode=False
    )

    test_results = pupyl_non_extreme.search(
        TEST_QUERY_IMAGE,
        top=expected_length_results
    )

    assert len([*test_results]) == expected_length_results


def test_search_non_extreme_mode_distances():
    """Unit test for method search, non-extreme mode and distances ase."""
    expected_result_index = 0
    expected_result_distance = .1

    pupyl_non_extreme = PupylImageSearch(
        data_dir=TEST_DATA_DIR,
        extreme_mode=False
    )

    test_results = pupyl_non_extreme.search(
        TEST_QUERY_IMAGE,
        top=1,
        return_distances=True,
        return_metadata=False
    )

    test_results = [*test_results][0]
    test_results_index = test_results[0]
    test_results_distance = round(test_results[1], 1)

    assert test_results_index == expected_result_index
    assert test_results_distance == expected_result_distance


def test_search_returning_metadata():
    """Unit test for method search, returning image metadata case."""
    with TemporaryDirectory() as temp_dir:
        test_pupyl = PupylImageSearch(temp_dir)
        test_pupyl.index(TEST_SCAN_DIR)

        test_results = [*test_pupyl.search(
            TEST_QUERY_IMAGE,
            top=1,
            return_metadata=True
        )][0]

        del test_results['original_access_time']

        assert isinstance(test_results, dict)


def test_remove():
    """Unit test for method remove."""
    index_to_remove = 0

    with TemporaryDirectory() as temp_dir:
        test_pupyl = PupylImageSearch(temp_dir)

        test_pupyl.index(TEST_SCAN_DIR)

        print([*os.listdir(os.path.join(temp_dir, '0'))])

        length_indexer_before = len(test_pupyl.indexer)
        length_image_database_before = len(test_pupyl.image_database)

        test_pupyl.remove(index_to_remove)

        length_indexer_after = len(test_pupyl.indexer)
        length_image_database_after = len(test_pupyl.image_database)

        assert length_indexer_after == length_indexer_before - 1
        assert length_image_database_after == length_image_database_before - 1


def test_remove_non_extreme():
    """Unit test for method remove, non extreme mode."""
    index_to_remove = 0

    with TemporaryDirectory() as temp_dir:
        pupyl_non_extreme = PupylImageSearch(
            data_dir=temp_dir,
            extreme_mode=False,
            characteristic=Characteristics.HEAVYWEIGHT_SLOW_GOOD_PRECISION
        )

        pupyl_non_extreme.index(TEST_SCAN_DIR)

        with Extractors(
            characteristics=Characteristics.HEAVYWEIGHT_SLOW_HUGE_PRECISION,
            extreme_mode=False
        ) as extractor:
            with Index(extractor.output_shape, data_dir=temp_dir) as indexer:
                length_indexer_before = len(indexer)
                length_image_database_before = len(
                    pupyl_non_extreme.image_database
                )

        pupyl_non_extreme.remove(index_to_remove)

        with Extractors(
            characteristics=Characteristics.HEAVYWEIGHT_SLOW_HUGE_PRECISION,
            extreme_mode=False
        ) as extractor:
            with Index(extractor.output_shape, data_dir=temp_dir) as indexer:
                length_indexer_after = len(indexer)
                length_image_database_after = len(
                    pupyl_non_extreme.image_database
                )

    assert length_image_database_after == \
        length_image_database_before - 1
    assert length_indexer_after == length_indexer_before - 1



"""Unit tests related to addendum.operators module."""
from pupyl.addendum.operators import intmul


TEST_VALUE = 123123
TEST_RESULT_ZERO_VALUE = 123
TEST_FACTOR = .001


def test_intmul_valid_case():
    """Unit test for method intmul, valid case."""
    assert TEST_VALUE << intmul >> TEST_FACTOR == 123


def test_intmul_below_zero_case():
    """Unit test for method intmul, below zero case."""
    assert TEST_RESULT_ZERO_VALUE << intmul >> TEST_FACTOR == 1



"""
Unit tests related to duplex.image module
"""
import os
import mimetypes
from unittest import TestCase

from pupyl.duplex.file_types import FileType, TarCompressedTypes
from pupyl.duplex.exceptions import FileTypeNotSupportedYet


TEST_DIR = 'tests'
TEST_LOCAL = os.path.join(TEST_DIR, 'test_image.jpg')
TEST_UNKNOWN = os.path.join(TEST_DIR, 'compressed_image.npy')
TEST_TAR_LOCATION = os.path.join(TEST_DIR, 'tar_files')


class TestCases(TestCase):
    """
    Unit tests over special cases
    """

    def test_guess_file_type_unknown_case(self):
        """
        Unit test for guess_file_type method, file type
        unknown case
        """
        with self.assertRaises(FileTypeNotSupportedYet):
            with open(TEST_UNKNOWN, 'rb') as ffile:
                FileType.guess_file_type(ffile)


def test_guess_file_type_known_case():
    """
    Unit test for guess_file_type method, file type
    known case
    """
    with open(TEST_LOCAL, 'rb') as ffile:
        assert FileType.guess_file_type(ffile.read()) == 'JPG'


def test_tar_compressed_types_resolve():
    """Unit test for tar_compressed_types_resolve method"""
    for ffile in os.listdir(TEST_TAR_LOCATION):
        test_type = mimetypes.guess_type(
            os.path.join(TEST_TAR_LOCATION, ffile)
        )[1]

        assert FileType.tar_compressed_types_resolve(
            test_type,
            mimetype=False
        ) == TarCompressedTypes.name(test_type)

        assert FileType.tar_compressed_types_resolve(
            test_type,
            mimetype=True
        ) == TarCompressedTypes.mime(test_type)
